# coding: utf-8

from supervisely_lib.project.project import Project
from supervisely_lib.annotation.tag_meta_collection import TagMetaCollection
from supervisely_lib.annotation.tag_meta import TagValueType
from supervisely_lib.sly_logger import logger


CLASSES_MAPPING = 'classes_mapping'
TAGS_MAPPING = 'tags_mapping'
CONFIDENCE_THRESHOLD = 'confidence_threshold'

TRUE_POSITIVE = 'true-positive'
TRUE_NEGATIVE = 'true-negative'
FALSE_POSITIVE = 'false-positive'
FALSE_NEGATIVE = 'false-negative'

ACCURACY = 'accuracy'
PRECISION = 'precision'
RECALL = 'recall'
F1_MEASURE = 'F1-measure'
TOTAL = 'total'
TOTAL_GROUND_TRUTH = 'total-ground-truth'
TOTAL_PREDICTIONS = 'total-predictions'

CONFUSION_MATRIX = 'confusion-matrix'
UNMATCHED_GT = 'unmatched-gt'
UNMATCHED_PREDICTIONS = 'unmatched-predictions'


def check_class_mapping(first_project: Project, second_project: Project, classes_mapping: dict) -> None:
    for k, v in classes_mapping.items():
        if first_project.meta.obj_classes.get(k) is None:
            raise RuntimeError('Class {} does not exist in input project "{}".'.format(k, first_project.name))
        if second_project.meta.obj_classes.get(v) is None:
            raise RuntimeError('Class {} does not exist in input project "{}".'.format(v, second_project.name))


def check_tag_mapping(first_project: Project, second_project: Project, tags_mapping: dict) -> None:
    for k, v in tags_mapping.items():
        if not first_project.meta.tag_metas.has_key(k):
            raise RuntimeError('Tag {} does not exist in input project "{}".'.format(k, first_project.name))
        if not second_project.meta.tag_metas.has_key(v):
            raise RuntimeError('Tag {} does not exist in input project "{}".'.format(v, second_project.name))


def render_labels_for_classes(labels, class_colors, canvas, missing_classes_color):
    for label in labels:
        color = class_colors.get(label.obj_class.name, missing_classes_color)
        label.geometry.draw(canvas, color)


def render_labels_for_class_name(labels, class_name, canvas):
    return render_labels_for_classes(labels, {class_name: True}, canvas, missing_classes_color=False)


def safe_ratio(num, denom):
    return (num / denom) if denom != 0 else 0


def sum_counters(elementwise_counters, counter_names):
    return {counter_name: sum(c.get(counter_name, 0) for c in elementwise_counters) for counter_name in counter_names}


def log_line(length=80, c=' '):
    logger.info(c * length)


def log_head(string):
    logger.info(string.center(80, '*'))



# coding: utf-8

import os
import time
from collections import defaultdict, OrderedDict
import json

from supervisely_lib.api.module_api import ApiField, ModuleApiBase, ModuleWithStatus, WaitingTimeExceeded
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from supervisely_lib.io.fs import get_file_name, ensure_base_path, get_file_hash
from supervisely_lib.collection.str_enum import StrEnum
from supervisely_lib._utils import batched


class TaskApi(ModuleApiBase, ModuleWithStatus):
    class RestartPolicy(StrEnum):
        NEVER = 'never'
        ON_ERROR = 'on_error'

    class PluginTaskType(StrEnum):
        TRAIN = 'train'
        INFERENCE = 'inference'
        INFERENCE_RPC = 'inference_rpc'
        SMART_TOOL = 'smarttool'
        CUSTOM = 'custom'

    class Status(StrEnum):
        QUEUED = 'queued'
        CONSUMED = 'consumed'
        STARTED = 'started'
        DEPLOYED = 'deployed'
        ERROR = 'error'
        FINISHED = 'finished'
        TERMINATING = 'terminating'
        STOPPED = 'stopped'

    def __init__(self, api):
        ModuleApiBase.__init__(self, api)
        ModuleWithStatus.__init__(self)

    def get_list(self, workspace_id, filters=None):
        '''
        :param workspace_id: int
        :param filters: list
        :return: list of all tasks in given workspace
        '''
        return self.get_list_all_pages('tasks.list',
                                       {ApiField.WORKSPACE_ID: workspace_id, ApiField.FILTER: filters or []})

    def get_info_by_id(self, id):
        '''
        :param id: int
        :return: tast metadata by numeric id
        '''
        return self._get_info_by_id(id, 'tasks.info')

    def get_status(self, task_id):
        '''
        :param task_id: int
        :return: Status class object (status of task with given id)
        '''
        status_str = self.get_info_by_id(task_id)[ApiField.STATUS]  # @TODO: convert json to tuple
        return self.Status(status_str)

    def raise_for_status(self, status):
        '''
        Raise error if status is ERROR
        :param status: Status class object
        '''
        if status is self.Status.ERROR:
            raise RuntimeError('Task status: ERROR')

    def wait(self, id, target_status, wait_attempts=None, wait_attempt_timeout_sec=None):
        '''
        Awaiting achievement by given task of a given status
        :param id: int
        :param target_status: Status class object (status of task we expect to destinate)
        :param wait_attempts: int
        :param wait_attempt_timeout_sec: int (raise error if waiting time exceeded)
        :return: bool
        '''
        wait_attempts = wait_attempts or self.MAX_WAIT_ATTEMPTS
        effective_wait_timeout = wait_attempt_timeout_sec or self.WAIT_ATTEMPT_TIMEOUT_SEC
        for attempt in range(wait_attempts):
            status = self.get_status(id)
            self.raise_for_status(status)
            if status is target_status:
                return
            time.sleep(effective_wait_timeout)
        raise WaitingTimeExceeded('Waiting time exceeded')

    def upload_dtl_archive(self, task_id, archive_path, progress_cb=None):
        encoder = MultipartEncoder({'id': str(task_id).encode('utf-8'),
                                    'name': get_file_name(archive_path),
                                    'archive': (
                                    os.path.basename(archive_path), open(archive_path, 'rb'), 'application/x-tar')})

        def callback(monitor_instance):
            read_mb = monitor_instance.bytes_read / 1024.0 / 1024.0
            if progress_cb is not None:
                progress_cb(read_mb)

        monitor = MultipartEncoderMonitor(encoder, callback)
        self._api.post('tasks.upload.dtl_archive', monitor)

    def _deploy_model(self, agent_id, model_id, plugin_id=None, version=None, restart_policy=RestartPolicy.NEVER,
                      settings=None):
        response = self._api.post('tasks.run.deploy', {ApiField.AGENT_ID: agent_id,
                                                       ApiField.MODEL_ID: model_id,
                                                       ApiField.RESTART_POLICY: restart_policy.value,
                                                       ApiField.SETTINGS: settings or {'gpu_device': 0},
                                                       ApiField.PLUGIN_ID: plugin_id,
                                                       ApiField.VERSION: version})
        return response.json()[ApiField.TASK_ID]

    def get_context(self, id):
        response = self._api.post('GetTaskContext', {ApiField.ID: id})
        return response.json()

    def _convert_json_info(self, info: dict):
        return info

    def run_dtl(self, workspace_id, dtl_graph, agent_id=None):
        response = self._api.post('tasks.run.dtl', {ApiField.WORKSPACE_ID: workspace_id,
                                                    ApiField.CONFIG: dtl_graph,
                                                    'advanced': {ApiField.AGENT_ID: agent_id}})
        return response.json()[ApiField.TASK_ID]

    def _run_plugin_task(self, task_type, agent_id, plugin_id, version, config, input_projects, input_models,
                         result_name):
        response = self._api.post('tasks.run.plugin', {'taskType': task_type,
                                                       ApiField.AGENT_ID: agent_id,
                                                       ApiField.PLUGIN_ID: plugin_id,
                                                       ApiField.VERSION: version,
                                                       ApiField.CONFIG: config,
                                                       'projects': input_projects,
                                                       'models': input_models,
                                                       ApiField.NAME: result_name})
        return response.json()[ApiField.TASK_ID]

    def run_train(self, agent_id, input_project_id, input_model_id, result_nn_name, train_config=None):
        model_info = self._api.model.get_info_by_id(input_model_id)
        return self._run_plugin_task(task_type=TaskApi.PluginTaskType.TRAIN.value,
                                     agent_id=agent_id,
                                     plugin_id=model_info.plugin_id,
                                     version=None,
                                     input_projects=[input_project_id],
                                     input_models=[input_model_id],
                                     result_name=result_nn_name,
                                     config={} if train_config is None else train_config)

    def run_inference(self, agent_id, input_project_id, input_model_id, result_project_name, inference_config=None):
        model_info = self._api.model.get_info_by_id(input_model_id)
        return self._run_plugin_task(task_type=TaskApi.PluginTaskType.INFERENCE.value,
                                     agent_id=agent_id,
                                     plugin_id=model_info.plugin_id,
                                     version=None,
                                     input_projects=[input_project_id],
                                     input_models=[input_model_id],
                                     result_name=result_project_name,
                                     config={} if inference_config is None else inference_config)

    def get_training_metrics(self, task_id):
        response = self._get_response_by_id(id=task_id, method='tasks.train-metrics', id_field=ApiField.TASK_ID)
        return response.json() if (response is not None) else None

    def deploy_model(self, agent_id, model_id):
        task_ids = self._api.model.get_deploy_tasks(model_id)
        if len(task_ids) == 0:
            task_id = self._deploy_model(agent_id, model_id)
        else:
            task_id = task_ids[0]
        self.wait(task_id, self.Status.DEPLOYED)
        return task_id

    def deploy_model_async(self, agent_id, model_id):
        task_ids = self._api.model.get_deploy_tasks(model_id)
        if len(task_ids) == 0:
            task_id = self._deploy_model(agent_id, model_id)
        else:
            task_id = task_ids[0]
        return task_id

    def stop(self, id):
        response = self._api.post('tasks.stop', {ApiField.ID: id})
        return self.Status(response.json()[ApiField.STATUS])

    def get_import_files_list(self, id):
        response = self._api.post('tasks.import.files_list', {ApiField.ID: id})
        return response.json() if (response is not None) else None

    def download_import_file(self, id, file_path, save_path):
        response = self._api.post('tasks.import.download_file', {ApiField.ID: id, ApiField.FILENAME: file_path}, stream=True)

        ensure_base_path(save_path)
        with open(save_path, 'wb') as fd:
            for chunk in response.iter_content(chunk_size=1024 * 1024):
                fd.write(chunk)

    def create_task_detached(self, workspace_id, task_type: str=None):
        response = self._api.post('tasks.run.python', {ApiField.WORKSPACE_ID: workspace_id,
                                                       ApiField.SCRIPT: "xxx",
                                                       ApiField.ADVANCED: {ApiField.IGNORE_AGENT: True}})
        return response.json()[ApiField.TASK_ID]

    def submit_logs(self, logs):
        response = self._api.post('tasks.logs.add', {ApiField.LOGS: logs})
        #return response.json()[ApiField.TASK_ID]

    def upload_files(self, task_id, abs_paths, names, progress_cb=None):
        if len(abs_paths) != len(names):
            raise RuntimeError("Inconsistency: len(abs_paths) != len(names)")

        hashes = []
        if len(abs_paths) == 0:
            return

        hash_to_items = defaultdict(list)
        hash_to_name = defaultdict(list)
        for idx, item in enumerate(zip(abs_paths, names)):
            path, name = item
            item_hash = get_file_hash(path)
            hashes.append(item_hash)
            hash_to_items[item_hash].append(path)
            hash_to_name[item_hash].append(name)

        unique_hashes = set(hashes)
        remote_hashes = self._api.image.check_existing_hashes(list(unique_hashes))
        new_hashes = unique_hashes - set(remote_hashes)

        # @TODO: upload remote hashes
        if len(remote_hashes) != 0:
            files = []
            for hash in remote_hashes:
                for name in hash_to_name[hash]:
                    files.append({ApiField.NAME: name, ApiField.HASH: hash})
            for batch in batched(files):
                resp = self._api.post('tasks.files.bulk.add-by-hash', {ApiField.TASK_ID: task_id, ApiField.FILES: batch})
        if progress_cb is not None:
            progress_cb(len(remote_hashes))

        for batch in batched(list(zip(abs_paths, names, hashes))):
            content_dict = OrderedDict()
            for idx, item in enumerate(batch):
                path, name, hash = item
                if hash in remote_hashes:
                    continue
                content_dict["{}".format(idx)] = json.dumps({"fullpath": name, "hash": hash})
                content_dict["{}-file".format(idx)] = (name, open(path, 'rb'), '')

            if len(content_dict) > 0:
                encoder = MultipartEncoder(fields=content_dict)
                resp = self._api.post('tasks.files.bulk.upload', encoder)
                if progress_cb is not None:
                    progress_cb(len(content_dict))

    # {
    #     data: {my_val: 1}
    #     obj: {val: 1, res: 2}
    # }
    # {
    #     obj: {new_val: 1}
    # }
    # // apped: true, recursive: false
    # {
    #     data: {my_val: 1}
    #     obj: {new_val: 1}
    # }(edited)
    # // append: false, recursive: false
    # {
    #     obj: {new_val: 1}
    # }(edited)
    #
    # 16: 32
    # // append: true, recursive: true
    # {
    #     data: {my_val: 1}
    #     obj: {val: 1, res: 2, new_val: 1}
    # }

    def set_fields(self, task_id, fields):
        for idx, obj in enumerate(fields):
            for key in [ApiField.FIELD, ApiField.PAYLOAD]:
                if key not in obj:
                    raise KeyError("Object #{} does not have field {!r}".format(idx, key))
        data = {
            ApiField.TASK_ID: task_id,
            ApiField.FIELDS: fields
        }
        resp = self._api.post('tasks.data.set', data)
        return resp.json()

    def set_field(self, task_id, field, payload, append=False, recursive=False):
        fields = [
            {
                ApiField.FIELD: field,
                ApiField.PAYLOAD: payload,
                ApiField.APPEND: append,
                ApiField.RECURSIVE: recursive,
            }
        ]
        return self.set_fields(task_id, fields)

    def get_fields(self, task_id, fields: list) -> dict:
        data = {
            ApiField.TASK_ID: task_id,
            ApiField.FIELDS: fields
        }
        resp = self._api.post('tasks.data.get', data)
        return resp.json()["result"]

    def get_field(self, task_id, field):
        result = self.get_fields(task_id, [field])
        return result[field]

    def _validate_checkpoints_support(self, task_id):
        info = self.get_info_by_id(task_id)
        if info["type"] != str(TaskApi.PluginTaskType.TRAIN):
            raise RuntimeError("Task (id={!r}) has type {!r}. "
                               "Checkpoints are available only for tasks of type {!r}".format())

    def list_checkpoints(self, task_id):
        self._validate_checkpoints_support(task_id)
        resp = self._api.post('tasks.checkpoints.list', {ApiField.ID: task_id})
        return resp.json()

    def delete_unused_checkpoints(self, task_id):
        self._validate_checkpoints_support(task_id)
        resp = self._api.post("tasks.checkpoints.clear", {ApiField.ID: task_id})
        return resp.json()

    def _set_output(self):
        pass

    def set_output_project(self, task_id, project_id, project_name=None):
        if project_name is None:
            project = self._api.project.get_info_by_id(project_id)
            project_name = project.name

        output = {
            ApiField.PROJECT: {
                ApiField.ID: project_id,
                ApiField.TITLE: project_name
            }
        }
        resp = self._api.post("tasks.output.set", {ApiField.TASK_ID: task_id, ApiField.OUTPUT: output})
        return resp.json()

    def set_output_report(self, task_id, file_id, file_name):
        return self._set_custom_output(task_id, file_id, file_name, description="Report", icon="zmdi zmdi-receipt")

    def _set_custom_output(self, task_id, file_id, file_name, file_url=None, description="File",
                           icon="zmdi zmdi-file-text", color="#33c94c", background_color="#d9f7e4",
                           download=False):
        if file_url is None:
            file_url = self._api.file.get_url(file_id)

        output = {
            ApiField.GENERAL: {
                "icon": {
                    "className": icon,
                    "color": color,
                    "backgroundColor": background_color
                },
                "title": file_name,
                "titleUrl": file_url,
                "download": download,
                "description": description
            }
        }
        resp = self._api.post("tasks.output.set", {ApiField.TASK_ID: task_id, ApiField.OUTPUT: output})
        return resp.json()

    def set_output_archive(self, task_id, file_id, file_name, file_url=None):
        if file_url is None:
            file_url = self._api.file.get_info_by_id(file_id).full_storage_url
        return self._set_custom_output(task_id, file_id, file_name,
                                       file_url=file_url,
                                       description="Download archive", icon="zmdi zmdi-archive",
                                       download=True)

    def send_request(self, task_id, method, data, context={}, skip_response=False):
        if type(data) is not dict:
            raise TypeError("data argument has to be a dict")
        resp = self._api.post("tasks.request.direct", {ApiField.TASK_ID: task_id,
                                                       ApiField.COMMAND: method,
                                                       ApiField.CONTEXT: context,
                                                       ApiField.STATE: data,
                                                       'skipResponse': skip_response})
        return resp.json()

    def set_output_directory(self, task_id, file_id, directory_path):
        return self._set_custom_output(task_id, file_id, directory_path, description="Directory", icon="zmdi zmdi-folder")


# coding: utf-8
from enum import IntEnum
from supervisely_lib.api.module_api import ApiField, ModuleApiBase


class RoleApi(ModuleApiBase):
    class DefaultRole(IntEnum):
        ADMIN =     1
        DEVELOPER = 2
        ANNOTATOR = 3
        VIEWER =    4

    @staticmethod
    def info_sequence():
        return [ApiField.ID,
                ApiField.ROLE,
                ApiField.CREATED_AT,
                ApiField.UPDATED_AT]

    @staticmethod
    def info_tuple_name():
        return 'RoleInfo'

    def get_list(self, filters=None):
        '''
        Print all roles that are available on private Supervisely instance
        :param filters: list
        :return: list
        '''
        return self.get_list_all_pages('roles.list', {ApiField.FILTER: filters or []})



# # coding: utf-8
#
# import os
# import os.path as osp
# import hashlib
# import shutil
#
# import supervisely_lib.io.fs as sly_fs
#
#
# # stateless object
# class FSStorage:
#     def __init__(self, name, storage_root):
#         self._name = name
#         self._storage_root = storage_root
#
#     @classmethod
#     def _copy_file_concurr(cls, src_path, dst_path):
#         try:
#             sly_fs.copy_file(src_path, dst_path)
#         except OSError as e:
#             # may be written by parallel process, skip in that case
#             if not osp.isfile(dst_path):
#                 raise e
#
#     @classmethod
#     def _copy_dir_recursively(cls, src_path, dst_path):
#         if sly_fs.dir_exists(dst_path):
#             sly_fs.remove_dir(dst_path)
#         #sly.fs.ensure_base_path(dst_path)
#         shutil.copytree(src_path, dst_path)
#
#     @ classmethod
#     def _get_obj_suffix(cls, st_path):
#         _, suffix = osp.splitext(st_path.rstrip('/'))
#         return suffix
#
#     def _storage_obj_exists(self, st_path, suffix):
#         raise NotImplementedError()
#
#     def _get_suffix(self, path):
#         raise NotImplementedError()
#
#     def _write_obj_impl(self, src_path, st_path):
#         raise NotImplementedError()
#
#     def _read_obj_impl(self, st_path, dst_path):
#         raise NotImplementedError()
#
#     def _rm_obj_impl(self, st_path):
#         raise NotImplementedError()
#
#     @property
#     def storage_root_path(self):
#         return self._storage_root
#
#     # slow
#     def list_objects(self):
#         def scan_deeper(paths):
#             res = []
#             for left_part in paths:
#                 for right_part in os.listdir(left_part):
#                     new_p = osp.join(left_part, right_part)
#                     res.append(new_p)
#             return res
#
#         obj_paths = [self._storage_root]
#         for lvl in range(3):  # fixed, determined by get_storage_path impl
#             obj_paths = scan_deeper(obj_paths)
#
#         obj_pathes_suffixes = [(p, self._get_suffix(p)) for p in obj_paths]
#         return obj_pathes_suffixes
#
#     def get_storage_path(self, data_hash, suffix=''):
#         if suffix:
#             suffix = ".{}".format(suffix).replace("..", ".")
#         st_hash = hashlib.sha256(data_hash.encode('utf-8')).hexdigest()
#         st_path = osp.join(self._storage_root, st_hash[0:2], st_hash[2:5], st_hash + suffix)
#         return st_path
#
#     def check_storage_object(self, data_hash, suffix=''):
#         st_path = self.get_storage_path(data_hash, suffix)
#         if self._storage_obj_exists(st_path, suffix):
#             return st_path
#         return None
#
#     def write_object(self, src_path, data_hash):
#         suffix = self._get_suffix(src_path)
#         st_path = self.get_storage_path(data_hash, suffix)
#         if not self._storage_obj_exists(st_path, suffix):
#             self._write_obj_impl(src_path, st_path)
#
#     def write_objects(self, src_paths, datas_hashes, progress_cb=None):
#         for src_path, data_hash in zip(src_paths, datas_hashes):
#             self.write_object(src_path, data_hash)
#             if progress_cb is not None:
#                 progress_cb()
#
#     def read_object(self, data_hash, dst_path):
#         suffix = self._get_suffix(dst_path)
#         st_path = self.check_storage_object(data_hash, suffix)
#         if not st_path:
#             return None  # doesn't exist
#         self._read_obj_impl(st_path, dst_path)
#         return dst_path
#
#     # @TODO: to remove??? used only in data_manager->download project
#     def read_objects(self, dst_paths_hashes, progress_ctr):
#         written_paths = []
#         for dst_path, data_hash in dst_paths_hashes:
#             res_path = self.read_object(data_hash, dst_path)
#             if res_path:
#                 written_paths.append(res_path)
#                 progress_ctr.iter_done_report()
#         return written_paths
#
#     def remove_object(self, st_path, suffix=''):
#         removable = self._storage_obj_exists(st_path, suffix)
#         if removable:
#             self._rm_obj_impl(st_path)
#         return removable
#
#
# class ImageStorage(FSStorage):
#     def _storage_obj_exists(self, st_path, suffix):
#         if not suffix:
#             raise ValueError('Storage {}. Image ext is empty.'.format(self._name))
#         return osp.isfile(st_path)
#
#     def _get_suffix(self, path):
#         return sly_fs.get_file_ext(path)
#
#     def _write_obj_impl(self, src_path, st_path):
#         sly_fs.ensure_base_path(st_path)
#         self._copy_file_concurr(src_path, st_path)
#
#     def _read_obj_impl(self, st_path, dst_path):
#         sly_fs.ensure_base_path(dst_path)
#         sly_fs.copy_file(st_path, dst_path)
#
#     def _rm_obj_impl(self, st_path):
#         os.remove(st_path)
#
#
# class NNStorage(FSStorage):
#     def _storage_obj_exists(self, st_path, suffix):
#         if suffix:
#             raise ValueError('Storage {}. Unexpected suffix for NN dir.'.format(self._name))
#         return osp.isdir(st_path)
#
#     def _get_suffix(self, path):
#         return ''
#
#     def _write_obj_impl(self, src_path, st_path):
#         self._copy_dir_recursively(src_path, st_path)
#
#     def _read_obj_impl(self, st_path, dst_path):
#         self._copy_dir_recursively(st_path, dst_path)
#
#     def _rm_obj_impl(self, st_path):
#         sly_fs.remove_dir(st_path)
#
#
# # class AnnStorage(FSStorage):
# #     def _storage_obj_exists(self, st_path, suffix):
# #         if not suffix:
# #             raise ValueError('Storage {}. Annotation ext is empty.'.format(self._name))
# #         return osp.isfile(st_path)
# #
# #     def _get_suffix(self, path):
# #         return '.json' #sly_fs.get_file_ext(path)
# #
# #     def _write_obj_impl(self, src_path, st_path):
# #         sly_fs.ensure_base_path(st_path)
# #         self._copy_file_concurr(src_path, st_path)
# #
# #     def _read_obj_impl(self, st_path, dst_path):
# #         sly_fs.ensure_base_path(dst_path)
# #         sly_fs.copy_file(st_path, dst_path)
# #
# #     def _rm_obj_impl(self, st_path):
# #         os.remove(st_path)
#
#
# class EmptyStorage(FSStorage):
#     def _storage_obj_exists(self, st_path, suffix):
#         return False
#
#     def _get_suffix(self, path):
#         return ''
#
#     def _write_obj_impl(self, src_path, st_path):
#         raise NotImplementedError()
#
#     def _read_obj_impl(self, st_path, dst_path):
#         raise NotImplementedError()
#
#     def _rm_obj_impl(self, st_path):
#         raise NotImplementedError()
#
#     def write_objects(self, src_paths, datas_hashes, progress_cb=None):
#         raise NotImplementedError()
#
#     def read_objects(self, dst_paths_hashes, progress_ctr):
#         return []  # overridden to speed up



# coding: utf-8
from supervisely_lib.video_annotation.video_object import VideoObject
from supervisely_lib.collection.key_indexed_collection import KeyIndexedCollection
from supervisely_lib.project.project_meta import ProjectMeta


class VideoObjectCollection(KeyIndexedCollection):
    '''
    Collection that stores VideoObject instances.
    '''
    item_type = VideoObject

    def to_json(self, key_id_map=None):
        '''
        Converts collection to json serializable list.
        :param key_id_map: KeyIdMap class object
        :return: list of VideoObjects in json format
        '''
        return [item.to_json(key_id_map) for item in self]

    @classmethod
    def from_json(cls, data, project_meta: ProjectMeta, key_id_map=None):
        '''
        Creates collection from json serializable list
        :param data: list of VideoObjects in json format
        :param project_meta: ProjectMeta class object
        :param key_id_map: KeyIdMap class object
        :return: VideoObjectCollection class object
        '''
        objects = [cls.item_type.from_json(video_object_json, project_meta, key_id_map) for video_object_json in data]
        return cls(objects)

    def __str__(self):
        return 'Objects:\n' + super(VideoObjectCollection, self).__str__()


# coding: utf-8

import cv2

from supervisely_lib.geometry.point_location import PointLocation
from supervisely_lib.geometry.geometry import Geometry
from supervisely_lib.geometry.rectangle import Rectangle
from supervisely_lib._utils import unwrap_if_numpy
from supervisely_lib.geometry.constants import LABELER_LOGIN, UPDATED_AT, CREATED_AT, ID, CLASS_ID


class Point(Geometry):
    def __init__(self, row, col,
                 sly_id=None, class_id=None, labeler_login=None, updated_at=None, created_at=None):
        """
        Create geopmetry point in (row, col) position. Float-type coordinates will be deprecated soon.
        """
        super().__init__(sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)
        self._row = round(unwrap_if_numpy(row))
        self._col = round(unwrap_if_numpy(col))

    @property
    def row(self):
        return self._row

    @property
    def col(self):
        return self._col

    @classmethod
    def from_point_location(cls, pt: PointLocation, sly_id=None, class_id=None, labeler_login=None, updated_at=None, created_at=None):
        '''
        The function from_point_location create Point class object from given PointLocation class object
        :param pt: PointLocation class object
        :return: Point class object
        '''
        return cls(row=pt.row, col=pt.col,
                   sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

    @property
    def point_location(self) -> PointLocation:
        '''
        The function point_location create PointLocation class object from Point class object
        :return: PointLocation class object
        '''
        return PointLocation(row=self.row, col=self.col)

    @staticmethod
    def geometry_name():
        return 'point'

    def crop(self, rect):
        '''
        :param rect: Rectangle class object
        :return: A copy of an object of class Point if this Rectangle contains Point and an empty list otherwise
        '''
        return [self.clone()] if rect.contains_point_location(self.point_location) else []

    def rotate(self, rotator):
        '''
        The function rotate calculates new parameters of point after rotating
        :param rotator: ImageRotator class object
        :return: Point class object
        '''
        return self.from_point_location(self.point_location.rotate(rotator))

    def resize(self, in_size, out_size):
        '''
        The function resize calculates new parameters of point after resizing image
        :param in_size: input image size
        :param out_size: output image size
        :return: Point class object
        '''
        return self.from_point_location(self.point_location.resize(in_size, out_size))

    def fliplr(self, img_size):
        '''
        The function fliplr calculates new parameters of point after fliping image in horizontal
        :param img_size: image size
        :return: Point class object
        '''
        return self.from_point_location(self.point_location.fliplr(img_size))

    def flipud(self, img_size):
        '''
        The function flipud calculates new parameters of point after fliping image in vertical
        :param img_size: image size
        :return: Point class object
        '''
        return self.from_point_location(self.point_location.flipud(img_size))

    def scale(self, factor):
        '''
        The function scale calculates new parameters of point after scaling image
        :param factor: float scale parameter
        :return: Point class object
        '''
        return self.from_point_location(self.point_location.scale(factor))

    def translate(self, drow, dcol):
        '''
        The function translate calculates new parameters of point after shifts it by a certain number of pixels
        :param drow: horizontal shift
        :param dcol: vertical shift
        :return: Point class object
        '''
        return self.from_point_location(self.point_location.translate(drow, dcol))

    def _draw_impl(self, bitmap, color, thickness=1, config=None):
        r = round(thickness / 2)  # @TODO: relation between thickness and point radius - ???
        cv2.circle(bitmap, (self.col, self.row), radius=r, color=color, thickness=cv2.FILLED)

    def _draw_contour_impl(self, bitmap, color, thickness=1, config=None):
        # @TODO: mb dummy operation for Point
        r = round((thickness + 1) / 2)
        cv2.circle(bitmap, (self.col, self.row), radius=r, color=color, thickness=cv2.FILLED)

    @property
    def area(self):
        return 0.0

    def to_bbox(self):
        '''
        The function to_bbox create Rectangle class object from Point class object
        :return: Rectangle class object
        '''
        return Rectangle(top=self.row, left=self.col, bottom=self.row, right=self.col)

    def to_json(self):
        '''
        The function to_json convert Point class object to json format
        :return: Point in json format
        '''
        res = self.point_location.to_json()
        self._add_creation_info(res)
        return res

    @classmethod
    def from_json(cls, data):
        '''
        The function from_json convert Point from json format to Point class object.
        :param data: input Point in json format
        :return: Point class object
        '''
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)
        return cls.from_point_location(PointLocation.from_json(data),
                                       sly_id=sly_id, class_id=class_id,
                                       labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

    @classmethod
    def allowed_transforms(cls):
        from supervisely_lib.geometry.any_geometry import AnyGeometry
        return [AnyGeometry]


# coding: utf-8
import uuid

# TODO: can not import from supervisely_lib.api.api import Api
from supervisely_lib.api.module_api import ApiField
from supervisely_lib.app_widget.widget_type import WidgetType
from supervisely_lib.app_widget.base_widget import BaseWidget


class InputWidget(BaseWidget):
    widget_type = WidgetType.FORM

    @classmethod
    def get_value(cls, api, app_id, widget_id, allow_default=True, value_type=None):
        data = api.report.get_widget(report_id=app_id, widget_id=widget_id)
        cls._validate_type(data)
        value = data["content"].get("value", None)
        if value is None and allow_default is True:
            value = data["content"].get("defaultValue", None)
        if value_type is not None:
            value = value_type(value)
        return value

    @classmethod
    def create(cls, name, description, id=None, default_value=None):
        res = dict()
        res[ApiField.TYPE] = str(cls.widget_type)
        res[ApiField.NAME] = "input block title"
        res[ApiField.SUBTITLE] = "input block description"
        res[ApiField.CONTENT] = {
            ApiField.NAME: name,
            ApiField.DESCRIPTION: description
        }
        res[ApiField.ID] = uuid.uuid4().hex if id is None else id
        if default_value is not None:
            res[ApiField.CONTENT][ApiField.DEFAULT_VALUE] = default_value

        return res



# coding: utf-8

from supervisely_lib.annotation.tag_collection import TagCollection
from supervisely_lib.annotation.obj_class import ObjClass
from supervisely_lib.geometry.any_geometry import AnyGeometry
from supervisely_lib.annotation.tag import Tag
from supervisely_lib.geometry.geometry import Geometry
from supervisely_lib.geometry.multichannel_bitmap import MultichannelBitmap
from supervisely_lib.imaging import image as sly_image
from supervisely_lib.project.project_meta import ProjectMeta
from supervisely_lib._utils import take_with_default
from supervisely_lib.annotation.json_geometries_map import GET_GEOMETRY_FROM_STR
from supervisely_lib.geometry.constants import GEOMETRY_TYPE, GEOMETRY_SHAPE


class LabelJsonFields:
    OBJ_CLASS_NAME = 'classTitle'
    DESCRIPTION = 'description'
    TAGS = 'tags'


class LabelBase:
    '''
    This is a class for creating and using labeling objects for annotations

    Attributes:
        geometry: Geometry class of the object(point, rectangle, polygon, bitmap, line)
        obj_class: Class of objects (person, car, etc) with necessary properties: name, type of geometry (Polygon, Rectangle, ...)
                   and RGB color. Only one class can be associated with Label.
        tags: TagCollection object
        description(str): description of the label
    '''
    def __init__(self, geometry: Geometry, obj_class: ObjClass, tags: TagCollection = None, description: str = ""):
        '''
        :param geometry: Geometry class of the object(point, rectangle, polygon, bitmap, line)
        :param obj_class: Class of objects (person, car, etc) with necessary properties: name, type of geometry (Polygon, Rectangle, ...)
                   and RGB color. Only one class can be associated with Label.
        :param tags: TagCollection object
        :param description: description of the label
        '''
        self._geometry = geometry
        self._obj_class = obj_class
        self._tags = take_with_default(tags, TagCollection())
        self._description = description
        self._validate_geometry_type()
        self._validate_geometry()

    def _validate_geometry(self):
        '''
        The function checks the name of the Object for compliance.
        :return: generate ValueError error if name is mismatch
        '''
        self._geometry.validate(self._obj_class.geometry_type.geometry_name(), self.obj_class.geometry_config)

    def _validate_geometry_type(self):
        raise NotImplementedError()

    @property
    def obj_class(self):
        return self._obj_class

    @property
    def geometry(self):
        return self._geometry
      
    @property
    def description(self):
        return self._description

    @property
    def tags(self):
        return self._tags.clone()

    def to_json(self):
        '''
        The function to_json convert label to json format
        :return: Label in json format
        '''

        res = {
            LabelJsonFields.OBJ_CLASS_NAME: self.obj_class.name,
            LabelJsonFields.DESCRIPTION: self.description,
            LabelJsonFields.TAGS: self.tags.to_json(),
            ** self.geometry.to_json(),
            GEOMETRY_TYPE: self.geometry.geometry_name(),
            GEOMETRY_SHAPE: self.geometry.geometry_name(),
        }
        return res

    @classmethod
    def from_json(cls, data, project_meta: ProjectMeta):
        '''
        The function from_json convert Label from json format to Label class object. If there is no ObjClass from
        input json format in ProjectMeta, it generate RuntimeError error.
        :param data: input label in json format
        :param project_meta: ProjectMeta class object
        :return: Label class object
        '''
        obj_class_name = data[LabelJsonFields.OBJ_CLASS_NAME]
        obj_class = project_meta.get_obj_class(obj_class_name)
        if obj_class is None:
            raise RuntimeError(f'Failed to deserialize a Label object from JSON: label class name {obj_class_name!r} '
                               f'was not found in the given project meta.')

        if obj_class.geometry_type is AnyGeometry:
            geometry_type_actual = GET_GEOMETRY_FROM_STR(data[GEOMETRY_TYPE] if GEOMETRY_TYPE in data else data[GEOMETRY_SHAPE])
            geometry = geometry_type_actual.from_json(data)
        else:
            geometry = obj_class.geometry_type.from_json(data)

        return cls(geometry=geometry,
                   obj_class=obj_class,
                   tags=TagCollection.from_json(data[LabelJsonFields.TAGS], project_meta.tag_metas),
                   description=data.get(LabelJsonFields.DESCRIPTION, ""))

    def add_tag(self, tag: Tag):
        '''
        The function add_tag add tag to the current Label object and return the copy of the
        current Label object
        :param tag: TagCollection class object to be added
        :return: Label class object with the new list of the tags
        '''
        return self.clone(tags=self._tags.add(tag))

    def add_tags(self, tags: list):
        '''
        The function add_tag add tags to the current Label object and return the copy of the
        current Label object
        :param tags: list of the TagCollection class objects to be added
        :return: Label class object with the new list of the tags
        :param tags:
        :return:
        '''
        return self.clone(tags=self._tags.add_items(tags))

    def clone(self, geometry: Geometry = None, obj_class: ObjClass = None, tags: TagCollection = None,
              description: str = None):
        '''
        The function clone make copy of the Label class object
        :return: Label class object
        '''
        return self.__class__(geometry=take_with_default(geometry, self.geometry),
                              obj_class=take_with_default(obj_class, self.obj_class),
                              tags=take_with_default(tags, self.tags),
                              description=take_with_default(description, self.description))

    def crop(self, rect):
        '''
        The function crop the current geometry of Label
        :param rect: Rectangle class object
        :return: Label class object with new geometry
        '''
        if rect.contains(self.geometry.to_bbox()):
            return [self]
        else:
            # for compatibility of old slightly invalid annotations, some of them may be out of image bounds.
            # will correct it automatically
            result_geometries = self.geometry.crop(rect)
            if len(result_geometries) == 1:
                result_geometries[0]._copy_creation_info_inplace(self.geometry)
                return [self.clone(geometry=result_geometries[0])]
            else:
                return [self.clone(geometry=g) for g in self.geometry.crop(rect)]

    def relative_crop(self, rect):
        '''
        The function relative_crop crop the current geometry of Label
        :param rect: Rectangle class object
        :return: Label class object with new geometry
        '''
        return [self.clone(geometry=g) for g in self.geometry.relative_crop(rect)]

    def rotate(self, rotator):
        '''
        The function rotate Label geometry and return the copy of the current Label object
        :param rotator: ImageRotator class object
        :return: Label class object with new(rotated) geometry
        '''
        return self.clone(geometry=self.geometry.rotate(rotator))

    def resize(self, in_size, out_size):
        '''
        The function resize Label geometry and return the copy of the current Label object
        :return: Label class object with new geometry
        '''
        return self.clone(geometry=self.geometry.resize(in_size, out_size))

    def scale(self, factor):
        '''
        The function scale change scale of the current Label object with a given factor
        :param factor: float scale parameter
        :return: Label class object with new geometry
        '''
        return self.clone(geometry=self.geometry.scale(factor))

    def translate(self, drow, dcol):
        '''
        The function translate shifts the object by a certain number of pixels and return the copy of the current Label object
        :param drow: horizontal shift
        :param dcol: vertical shift
        :return: Label class object with new geometry
        '''
        return self.clone(geometry=self.geometry.translate(drow=drow, dcol=dcol))

    def fliplr(self, img_size):
        '''
        The function fliplr the current Label object geometry in horizontal and return the copy of the
        current Label object
        :param img_size: size of the image
        :return: Label class object with new geometry
        '''
        return self.clone(geometry=self.geometry.fliplr(img_size))

    def flipud(self, img_size):
        '''
        The function fliplr the current Label object geometry in vertical and return the copy of the
        current Label object
        :param img_size: size of the image
        :return: Label class object with new geometry
        '''
        return self.clone(geometry=self.geometry.flipud(img_size))

    def _draw_tags(self, bitmap, font):
        '''
        The function _draw_tags text of the tags on bitmap from left to right.
        :param bitmap: target image (np.ndarray)
        :param font: size of the font
        '''
        bbox = self.geometry.to_bbox()
        texts = [tag.get_compact_str() for tag in self.tags]
        sly_image.draw_text_sequence(bitmap=bitmap,
                                     texts=texts,
                                     anchor_point=(bbox.top, bbox.left),
                                     corner_snap=sly_image.CornerAnchorMode.BOTTOM_LEFT,
                                     font=font)

    def draw(self, bitmap, color=None, thickness=1, draw_tags=False, tags_font=None):
        '''
        The function draws rectangle near label geometry on bitmap
        :param bitmap: target image (np.ndarray)
        :param color: [R, G, B]
        :param thickness: thickness of the drawing rectangle
        '''
        effective_color = take_with_default(color, self.obj_class.color)
        self.geometry.draw(bitmap, effective_color, thickness, config=self.obj_class.geometry_config)
        if draw_tags:
            self._draw_tags(bitmap, tags_font)

    def draw_contour(self, bitmap, color=None, thickness=1, draw_tags=False, tags_font=None):
        '''
        The function draws the figure contour on a given label geometry bitmap
        :param bitmap: target image (np.ndarray)
        :param color: [R, G, B]
        :param thickness: thickness of the drawing contour
        '''
        effective_color = take_with_default(color, self.obj_class.color)
        self.geometry.draw_contour(bitmap, effective_color, thickness, config=self.obj_class.geometry_config)
        if draw_tags:
            self._draw_tags(bitmap, tags_font)

    @property
    def area(self):
        '''
        :return: area of current geometry in Label object
        '''
        return self.geometry.area

    def convert(self, new_obj_class: ObjClass):
        labels = []
        geometries = self.geometry.convert(new_obj_class.geometry_type)
        for g in geometries:
            labels.append(self.clone(geometry=g, obj_class=new_obj_class))
        return labels


class Label(LabelBase):
    def _validate_geometry_type(self):
        '''
        Checks geometry type for correctness
        '''
        if self._obj_class.geometry_type != AnyGeometry:
            if type(self._geometry) is not self._obj_class.geometry_type:
                raise RuntimeError("Input geometry type {!r} != geometry type of ObjClass {}"
                                   .format(type(self._geometry), self._obj_class.geometry_type))


class PixelwiseScoresLabel(LabelBase):
    def _validate_geometry_type(self):
        '''
        Checks geometry type for correctness
        '''
        if type(self._geometry) is not MultichannelBitmap:
            raise RuntimeError("Input geometry type {!r} != geometry type of ObjClass {}"
                               .format(type(self._geometry), MultichannelBitmap))


# coding: utf-8

from typing import List
from collections import defaultdict
from supervisely_lib.collection.key_indexed_collection import KeyIndexedCollection
from supervisely_lib.io.json import JsonSerializable
from supervisely_lib.annotation.obj_class import ObjClass
from supervisely_lib.imaging.color import rgb2hex, hex2rgb


class ObjClassCollection(KeyIndexedCollection, JsonSerializable):
    """
    Collection that stores ObjClass instances with unique names.
    It raises error if the name of added item already exists.
    """

    item_type = ObjClass

    def to_json(self) -> List[dict]:
        """
        Converts collection to json serializable list. See Supervisely Json format explanation here:
        https://docs.supervise.ly/ann_format/

        Returns:
            json serializable dictionary
        """
        return [obj_class.to_json() for obj_class in self]

    @classmethod
    def from_json(cls, data: List[dict]) -> 'ObjClassCollection':
        """
        Creates collection from json serializable list. See Supervisely Json format explanation here:
        https://docs.supervise.ly/ann_format/

        Returns:
            ObjClassCollection
        """
        obj_classes = [ObjClass.from_json(obj_class_json) for obj_class_json in data]
        return cls(obj_classes)

    def validate_classes_colors(self, logger=None):
        # check colors uniq
        color_names = defaultdict(list)
        for obj_class in self:
            hex = rgb2hex(obj_class.color)
            color_names[hex].append(obj_class.name)

        class_colors_notify = None
        for hex_color, class_names in color_names.items():
            if len(class_names) > 1:
                warn_str = "Classes {!r} have the same RGB color = {!r}".format(class_names, hex2rgb(hex_color))
                if logger is not None:
                    logger.warn(warn_str)
                if class_colors_notify is None:
                    class_colors_notify = ""
                class_colors_notify += warn_str + '\n\n'
        return class_colors_notify


def make_renamed_classes(src_obj_classes: ObjClassCollection, renamer, skip_missing=False) -> ObjClassCollection:
    renamed_classes = []
    for src_cls in src_obj_classes:
        renamed_name = renamer.rename(src_cls.name)
        if renamed_name is not None:
            renamed_classes.append(src_cls.clone(name=renamed_name))
        elif not skip_missing:
            raise KeyError('Object class name {} could not be mapped to a destination name.'.format(src_cls.name))
    return ObjClassCollection(items=renamed_classes)



# coding: utf-8

import itertools
from collections import defaultdict
from copy import deepcopy
import numpy as np
import pkg_resources

from supervisely_lib.metric.common import safe_ratio
from supervisely_lib.project.project_meta import ProjectMeta
from supervisely_lib.annotation.annotation import Annotation
from supervisely_lib.annotation.label import Label
from supervisely_lib.annotation.obj_class import ObjClass
from supervisely_lib.annotation.obj_class_collection import make_renamed_classes, ObjClassCollection
from supervisely_lib.annotation.obj_class_mapper import ObjClassMapper, RenamingObjClassMapper
from supervisely_lib.annotation.tag_meta_collection import make_renamed_tag_metas
from supervisely_lib.annotation.tag_meta import TagValueType
from supervisely_lib.annotation.tag_meta_mapper import RenamingTagMetaMapper, TagMetaMapper, make_renamed_tags
from supervisely_lib.annotation.renamer import Renamer, is_name_included
from supervisely_lib.geometry.geometry import Geometry
from supervisely_lib.geometry.multichannel_bitmap import MultichannelBitmap
from supervisely_lib.geometry.rectangle import Rectangle
from supervisely_lib.geometry.sliding_windows import SlidingWindows
from supervisely_lib.imaging.image import read as sly_image_read
from supervisely_lib.nn import raw_to_labels
from supervisely_lib.nn.config import update_recursively, update_strict, MultiTypeValidator
from supervisely_lib.nn.hosted.inference_single_image import SingleImageInferenceBase


INFERENCE_MODE_CONFIG = 'inference_mode_config'
CLASS_NAME = 'class_name'
FROM_CLASSES = 'from_classes'
MODEL_CLASSES = 'model_classes'
MODEL_TAGS = 'model_tags'
MODE = 'mode'
NAME = 'name'
SAVE = 'save'

MATCH_ALL = '__all__'

WINDOW = 'window'
HEIGHT = 'height'
WIDTH = 'width'
MIN_OVERLAP = 'min_overlap'
X = 'x'
Y = 'y'

NMS_AFTER = 'nms_after'
ENABLE = 'enable'
IOU_THRESHOLD = 'iou_threshold'

# TODO move somewhere near geometric transformations.
BOUNDS = 'bounds'
PADDING = 'padding'

BOTTOM = 'bottom'
LEFT = 'left'
RIGHT = 'right'
TOP = 'top'

PERCENT = '%'
PX = 'px'

CONFIDENCE = 'confidence'
CONFIDENCE_TAG_NAME = 'confidence_tag_name'

SAVE_PROBABILITIES = 'save_probabilities'


def _replace_or_drop_labels_classes(labels, obj_class_mapper: ObjClassMapper, tag_meta_mapper: TagMetaMapper) -> list:
    result = []
    for label in labels:
        dest_obj_class = obj_class_mapper.map(label.obj_class)
        if dest_obj_class is not None:
            renamed_tags = make_renamed_tags(tags=label.tags, tag_meta_mapper=tag_meta_mapper, skip_missing=True)
            result.append(label.clone(obj_class=dest_obj_class, tags=renamed_tags))
    return result


def _rectangle_from_cropping_or_padding_bounds(img_shape, crop_config, do_crop: bool):
    def get_crop_pixels(raw_side, dim_name):
        side_crop_config = crop_config.get(dim_name)
        if side_crop_config is None:
            crop_pixels = 0
        elif side_crop_config.endswith(PX):
            crop_pixels = int(side_crop_config[:-len(PX)])
        elif side_crop_config.endswith(PERCENT):
            padding_fraction = float(side_crop_config[:-len(PERCENT)])
            crop_pixels = int(raw_side * padding_fraction / 100.0)
        else:
            raise ValueError(
                'Unknown padding size format: {}. Expected absolute values as "5px" or relative as "5%"'.format(
                    side_crop_config))
        if not do_crop:
            crop_pixels *= -1  # Pad instead of crop.
        return crop_pixels

    # TODO more informative error message.
    return Rectangle(top=get_crop_pixels(img_shape[0], TOP),
                     left=get_crop_pixels(img_shape[1], LEFT),
                     bottom=img_shape[0] - get_crop_pixels(img_shape[0], BOTTOM) - 1,
                     right=img_shape[1] - get_crop_pixels(img_shape[1], RIGHT) - 1)


def _make_cropped_rectangle(img_shape, crop_config):
    return _rectangle_from_cropping_or_padding_bounds(img_shape, crop_config, do_crop=True)


def _make_padded_rectangle(img_shape, pad_config):
    return _rectangle_from_cropping_or_padding_bounds(img_shape, pad_config, do_crop=False)


def _is_backend_only_geometry(geom: Geometry) -> bool:
    """Checks whether a figure is backend-only (i.e. should not be saved as a part of final image annotation)."""
    return any(isinstance(geom, geom_type) for geom_type in [MultichannelBitmap])


def _remove_backend_only_labels(labels):
    return [label for label in labels if not _is_backend_only_geometry(label.geometry)]


def _maybe_make_intermediate_bbox_class(save_config):
    return None if not save_config.get(SAVE, False) else ObjClass(name=save_config[CLASS_NAME],
                                                                  geometry_type=Rectangle)


def _maybe_make_bbox_label(roi: Rectangle, bbox_class: ObjClass, tags=None) -> list:
    return [Label(geometry=roi, obj_class=bbox_class, tags=tags)] if bbox_class is not None else []


def _get_annotation_for_bbox(img: np.ndarray, roi: Rectangle, model) -> Annotation:
    """Runs inference within the given roi; moves resulting figures to global reference frame."""
    img_cropped = roi.get_cropped_numpy_slice(img)
    # TODO pass through image and parent figure tags via roi_ann.
    roi_ann = Annotation(img_size=(roi.height, roi.width))
    raw_result_ann = model.inference(img_cropped, roi_ann)
    return Annotation(img_size=img.shape[:2],
                      labels=[label.translate(drow=roi.top, dcol=roi.left) for label in raw_result_ann.labels],
                      img_tags=raw_result_ann.img_tags, img_description=raw_result_ann.img_description,
                      pixelwise_scores_labels=[label.translate(drow=roi.top, dcol=roi.left)
                                               for label in raw_result_ann.pixelwise_scores_labels])


class InferenceModeBase:
    @staticmethod
    def mode_name():
        raise NotImplementedError()

    @classmethod
    def make_default_config(cls, model_result_suffix):
        return {
            MODEL_CLASSES: Renamer(add_suffix=model_result_suffix, save_names=MATCH_ALL).to_json(),
            MODEL_TAGS: Renamer(add_suffix=model_result_suffix, save_names=MATCH_ALL).to_json(),
            NAME: cls.mode_name(),
            SAVE_PROBABILITIES: False
        }

    def __init__(self, config: dict, in_meta: ProjectMeta, model: SingleImageInferenceBase):
        validation_schema_path = pkg_resources.resource_filename(
            __name__, 'inference_modes_schemas/{}.json'.format(self.mode_name()))
        MultiTypeValidator(validation_schema_path).val(INFERENCE_MODE_CONFIG, config)
        self._config = deepcopy(config)
        self._out_meta = in_meta
        self._model = model
        model_out_meta = self._model.model_out_meta

        # Renamer for model classes.
        renamer_model = Renamer.from_json(config.get(MODEL_CLASSES, {}))
        # Add all the applicable (passing the renamer filter) renamed model classes to the output meta.
        self._out_meta = self.out_meta.add_obj_classes(
            make_renamed_classes(model_out_meta.obj_classes, renamer_model, skip_missing=True))
        # Make a class mapper to translate from model object space to output meta space.
        # TODO store the renamed model classes separately for the mapper instead of mixing in with the input annotation
        # classes.
        self._model_class_mapper = RenamingObjClassMapper(dest_obj_classes=self._out_meta.obj_classes,
                                                          renamer=renamer_model)

        # Renamer for model tags.
        self._model_tags_renamer = Renamer.from_json(config.get(MODEL_TAGS, {}))
        # Rename the model output tags, set up a mapper and add them to the output meta.
        self._renamed_model_tags = make_renamed_tag_metas(
            model_out_meta.tag_metas, self._model_tags_renamer, skip_missing=True)
        self._model_tag_meta_mapper = RenamingTagMetaMapper(dest_tag_meta_dict=self._renamed_model_tags,
                                                            renamer=self._model_tags_renamer)
        self._out_meta = self._out_meta.add_tag_metas(self._renamed_model_tags)

    @property
    def out_meta(self) -> ProjectMeta:
        return self._out_meta

    def _make_final_ann(self, result_ann):
        frontend_compatible_labels = _remove_backend_only_labels(result_ann.labels)
        return Annotation(img_size=result_ann.img_size,
                          labels=frontend_compatible_labels,
                          img_tags=result_ann.img_tags,
                          img_description=result_ann.img_description,
                          pixelwise_scores_labels=result_ann.pixelwise_scores_labels)

    def infer_annotate(self, img: np.ndarray, ann: Annotation):
        result_ann = self._do_infer_annotate(img, ann)
        return self._make_final_ann(result_ann)

    def infer_annotate_image_file(self, image_file: str, ann: Annotation):
        result_ann = self._do_infer_annotate_image_file(image_file, ann)
        return self._make_final_ann(result_ann)

    def _do_infer_annotate(self, img, ann: Annotation) -> Annotation:
        raise NotImplementedError()

    def _do_infer_annotate_image_file(self, image_file: str, ann: Annotation) -> Annotation:
        img = sly_image_read(image_file)
        return self._do_infer_annotate(img, ann)


class InfModeFullImage(InferenceModeBase):
    @staticmethod
    def mode_name():
        return 'full_image'

    def _do_infer_annotate_generic(self, inference_fn, img, ann: Annotation):
        result_ann = ann.clone()
        inference_ann = inference_fn(img, ann)

        result_labels = _replace_or_drop_labels_classes(
            inference_ann.labels, self._model_class_mapper, self._model_tag_meta_mapper)
        result_ann = result_ann.add_labels(result_labels)

        renamed_tags = make_renamed_tags(inference_ann.img_tags, self._model_tag_meta_mapper, skip_missing=True)
        result_ann = result_ann.add_tags(renamed_tags)

        if self._config.get(SAVE_PROBABILITIES, False) is True:
            result_problabels = _replace_or_drop_labels_classes(
                inference_ann.pixelwise_scores_labels, self._model_class_mapper, self._model_tag_meta_mapper)
            result_ann = result_ann.add_pixelwise_score_labels(result_problabels)

        return result_ann

    def _do_infer_annotate(self, img: np.ndarray, ann: Annotation) -> Annotation:
        return self._do_infer_annotate_generic(self._model.inference, img, ann)

    def _do_infer_annotate_image_file(self, image_file: str, ann: Annotation) -> Annotation:
        return self._do_infer_annotate_generic(self._model.inference_image_file, image_file, ann)


class InfModeRoi(InferenceModeBase):
    @staticmethod
    def mode_name():
        return 'roi'

    @classmethod
    def make_default_config(cls, model_result_suffix: str):
        config = super(InfModeRoi, cls).make_default_config(model_result_suffix)
        our_config = {
            BOUNDS: {
                LEFT: '0' + PX,
                TOP: '0' + PX,
                RIGHT: '0' + PX,
                BOTTOM: '0' + PX,
            },
            SAVE: False,
            CLASS_NAME: 'inference_roi'
        }
        update_strict(config, our_config)
        return config

    def __init__(self, config: dict, in_meta: ProjectMeta, model: SingleImageInferenceBase):
        super().__init__(config, in_meta, model)
        self._intermediate_bbox_class = _maybe_make_intermediate_bbox_class(self._config)
        if self._intermediate_bbox_class is not None:
            self._out_meta = self._out_meta.add_obj_class(self._intermediate_bbox_class)

    def _do_infer_annotate(self, img: np.ndarray, ann: Annotation) -> Annotation:
        result_ann = ann.clone()
        roi = _make_cropped_rectangle(ann.img_size, self._config[BOUNDS])
        roi_ann = _get_annotation_for_bbox(img, roi, self._model)
        result_ann = result_ann.add_labels(
            _replace_or_drop_labels_classes(roi_ann.labels, self._model_class_mapper, self._model_tag_meta_mapper))
        img_level_tags = make_renamed_tags(roi_ann.img_tags, self._model_tag_meta_mapper, skip_missing=True)
        result_ann = result_ann.add_labels(
            _maybe_make_bbox_label(roi, self._intermediate_bbox_class, tags=img_level_tags))
        result_ann = result_ann.add_tags(img_level_tags)

        if self._config.get(SAVE_PROBABILITIES, False) is True:
            result_problabels = _replace_or_drop_labels_classes(
                roi_ann.pixelwise_scores_labels, self._model_class_mapper, self._model_tag_meta_mapper)
            result_ann = result_ann.add_pixelwise_score_labels(result_problabels)

        return result_ann


class InfModeBboxes(InferenceModeBase):
    @staticmethod
    def mode_name():
        return 'bboxes'

    @classmethod
    def make_default_config(cls, model_result_suffix: str) -> dict:
        config = super(InfModeBboxes, cls).make_default_config(model_result_suffix)
        our_config = {
            FROM_CLASSES: MATCH_ALL,
            PADDING: {
                LEFT: '0' + PX,
                TOP: '0' + PX,
                RIGHT: '0' + PX,
                BOTTOM: '0' + PX,
            },
            SAVE: False,
            Renamer.ADD_SUFFIX: '_input_bbox'
        }
        update_strict(config, our_config)
        return config

    def __init__(self, config: dict, in_meta: ProjectMeta, model: SingleImageInferenceBase):
        super().__init__(config, in_meta, model)

        # If saving the bounding boxes on which inference was called is requested, create separate classes
        # for those bounding boxes by renaming the source object classes.
        self._renamer_intermediate = None
        if self._config[SAVE]:
            renamer_intermediate = Renamer(add_suffix=self._config[Renamer.ADD_SUFFIX],
                                           save_names=self._config[FROM_CLASSES])
            # First simply rename the matching source classes.
            intermediate_renamed_classes = make_renamed_classes(in_meta.obj_classes, renamer_intermediate,
                                                                skip_missing=True)
            # Next, change the geometry type for the intermediate bounding box classes to Rectangle.
            intermediate_renamed_rectangle_classes = ObjClassCollection(items=[
                renamed_class.clone(geometry_type=Rectangle) for renamed_class in intermediate_renamed_classes])
            # Add the renamed Rectangle classes to the output meta and set up a class mapper.
            self._out_meta = self._out_meta.add_obj_classes(intermediate_renamed_rectangle_classes)
            self._intermediate_class_mapper = RenamingObjClassMapper(
                dest_obj_classes=intermediate_renamed_rectangle_classes, renamer=renamer_intermediate)

    def _do_infer_annotate(self, img: np.ndarray, ann: Annotation) -> Annotation:
        result_labels = []
        result_problabels = []
        for src_label, roi in self._all_filtered_bbox_rois(ann, self._config[FROM_CLASSES], self._config[PADDING]):
            if roi is None:
                result_labels.append(src_label)
            else:
                roi_ann = _get_annotation_for_bbox(img, roi, self._model)
                result_labels.extend(_replace_or_drop_labels_classes(
                    roi_ann.labels, self._model_class_mapper, self._model_tag_meta_mapper))

                if self._config.get(SAVE_PROBABILITIES, False) is True:
                    result_problabels.extend(_replace_or_drop_labels_classes(roi_ann.pixelwise_scores_labels,
                                                                             self._model_class_mapper,
                                                                             self._model_tag_meta_mapper))

                model_img_level_tags = make_renamed_tags(roi_ann.img_tags, self._model_tag_meta_mapper,
                                                         skip_missing=True)
                if self._config[SAVE]:
                    result_labels.append(
                        Label(geometry=roi, obj_class=self._intermediate_class_mapper.map(src_label.obj_class),
                              tags=model_img_level_tags))
                # Regardless of whether we need to save intermediate bounding boxes, also put the inference result tags
                # onto the original source object from which we created a bounding box.
                # This is necessary for e.g. classification models to work, so that they put the classification results
                # onto the original object.
                result_labels.append(src_label.add_tags(model_img_level_tags))
        return ann.clone(labels=result_labels, pixelwise_scores_labels=result_problabels)

    @staticmethod
    def _all_filtered_bbox_rois(ann: Annotation, included_classes, crop_config: dict):
        for src_label in ann.labels:
            effective_roi = None
            if is_name_included(src_label.obj_class.name, included_classes):
                bbox = src_label.geometry.to_bbox()
                roi = _make_padded_rectangle((bbox.height, bbox.width), crop_config)
                maybe_effective_roi = roi.translate(drow=bbox.top, dcol=bbox.left).crop(
                    Rectangle.from_size(ann.img_size))
                if len(maybe_effective_roi) > 0:
                    [effective_roi] = maybe_effective_roi
            yield src_label, effective_roi


class InfModeSlidinglWindowBase(InferenceModeBase):
    def __init__(self, config: dict, in_meta: ProjectMeta, model: SingleImageInferenceBase):
        super().__init__(config, in_meta, model)

        window_shape = (self._config[WINDOW][HEIGHT], self._config[WINDOW][WIDTH])
        min_overlap = (self._config[MIN_OVERLAP][Y], self._config[MIN_OVERLAP][X])
        self._sliding_windows = SlidingWindows(window_shape, min_overlap)

        self._intermediate_bbox_class = _maybe_make_intermediate_bbox_class(self._config)
        if self._intermediate_bbox_class is not None:
            self._out_meta = self._out_meta.add_obj_class(self._intermediate_bbox_class)

    @classmethod
    def make_default_config(cls, model_result_suffix: str) -> dict:
        config = super(InfModeSlidinglWindowBase, cls).make_default_config(model_result_suffix)
        our_config = {
            WINDOW: {
                WIDTH: 128,
                HEIGHT: 128,
            },
            MIN_OVERLAP: {
                X: 0,
                Y: 0,
            },
            SAVE: False,
            CLASS_NAME: 'sliding_window',
        }
        update_strict(config, our_config)
        return config


# This only makes sense for image segmentation that return per-pixel class probabilities.
class InfModeSlidingWindowSegmentation(InfModeSlidinglWindowBase):
    @staticmethod
    def mode_name():
        return 'sliding_window'

    def _do_infer_annotate(self, img: np.ndarray, ann: Annotation) -> Annotation:
        result_ann = ann.clone()
        all_pixelwise_scores_labels = []
        for roi in self._sliding_windows.get(ann.img_size):
            raw_roi_ann = _get_annotation_for_bbox(img, roi, self._model)
            all_pixelwise_scores_labels.extend(raw_roi_ann.pixelwise_scores_labels)
            model_img_level_tags = make_renamed_tags(raw_roi_ann.img_tags, self._model_tag_meta_mapper,
                                                     make_renamed_tags)
            result_ann = result_ann.add_labels(
                _maybe_make_bbox_label(roi, self._intermediate_bbox_class, tags=model_img_level_tags))
        model_class_name_to_id = {name: idx
                                  for idx, name in enumerate(set(label.obj_class.name
                                                                 for label in all_pixelwise_scores_labels))}
        id_to_class_obj = {idx: self._model.model_out_meta.obj_classes.get(name)
                           for name, idx in model_class_name_to_id.items()}
        summed_scores = np.zeros(ann.img_size + tuple([len(model_class_name_to_id)]))
        summed_divisor = np.zeros_like(summed_scores)
        for label in all_pixelwise_scores_labels:
            class_idx = model_class_name_to_id[label.obj_class.name]
            geom_bbox = label.geometry.to_bbox()
            label_matching_summer_scores = geom_bbox.get_cropped_numpy_slice(summed_scores)
            label_matching_summer_scores[:, :, class_idx, np.newaxis] += label.geometry.data

            divisor_slice = geom_bbox.get_cropped_numpy_slice(summed_divisor)
            divisor_slice[:, :, class_idx, np.newaxis] += 1.

        # TODO consider instead filtering pixels by all-zero scores.
        if np.sum(summed_scores, axis=2).min() == 0:
            raise RuntimeError('Wrong sliding window moving, implementation error.')
        aggregated_model_labels = raw_to_labels.segmentation_array_to_sly_bitmaps(id_to_class_obj,
                                                                                  np.argmax(summed_scores, axis=2))
        result_ann = result_ann.add_labels(
            _replace_or_drop_labels_classes(
                aggregated_model_labels, self._model_class_mapper, self._model_tag_meta_mapper))

        if self._config.get(SAVE_PROBABILITIES, False) is True:
            # copied fom unet's inference.py
            mean_scores = summed_scores / summed_divisor
            accumulated_pixelwise_scores_labels = raw_to_labels.segmentation_scores_to_per_class_labels(
                id_to_class_obj, mean_scores)
            result_problabels = _replace_or_drop_labels_classes(
                accumulated_pixelwise_scores_labels, self._model_class_mapper, self._model_tag_meta_mapper)
            result_ann = result_ann.add_pixelwise_score_labels(result_problabels)

        return result_ann


class InfModeSlidingWindowDetection(InfModeSlidinglWindowBase):
    @staticmethod
    def mode_name():
        return 'sliding_window_det'

    @classmethod
    def make_default_config(cls, model_result_suffix: str) -> dict:
        config = super(InfModeSlidingWindowDetection, cls).make_default_config(model_result_suffix)
        our_config = {
            NMS_AFTER: {
                ENABLE: False,
                IOU_THRESHOLD: 0.0,
                CONFIDENCE_TAG_NAME: CONFIDENCE
            },
        }
        update_strict(config, our_config)
        return config

    @staticmethod
    def _iou(first_rect, second_rect):
        intersection_rects = first_rect.crop(second_rect)
        if len(intersection_rects) == 0:
            return 0
        intersection_rect = intersection_rects[0]
        return safe_ratio(intersection_rect.area, first_rect.area + second_rect.area - intersection_rect.area)

    @classmethod
    def _single_class_nms(cls, labels, iou_thresh, confidence_tag_name):
        # We have to sort in the order of increasing score and check *all* the labels with a higher confidence than our
        # given label to make sure to filter out low-confidence labels transitively.
        #
        # I.e., if we have A > B > C (by confidence) and (A intersects B), (B intersects C), but (A not intersects C),
        # and we iteratively filter from the highest confidence first, then on step 1 we will only filter out B (because
        # A does not intersect C, so C remains), and on step 2 we will not filter out C because B is already gone, so
        # the end result will be [A, C]
        #
        # If we start from the bottom though, then we will first filter out C (by looking at B) and then filter out B
        # (by looking at A), so the end result will be only [A], which is what we want.
        sorted_labels = sorted(labels, key=lambda x: x.tags.get(confidence_tag_name).value)
        return [
            label for label_idx, label in enumerate(sorted_labels)
            if all(cls._iou(label.geometry.to_bbox(), other.geometry.to_bbox()) <= iou_thresh
                   for other in sorted_labels[(label_idx + 1):])]

    # @TODO: move out
    @classmethod
    def _general_nms(cls, labels, iou_thresh, confidence_tag_name):
        if not all(isinstance(label.geometry, Rectangle) for label in labels):
            raise RuntimeError('Non-max suppression expects labels with Rectangle geometry.')
        if not all(label.tags.has_key(confidence_tag_name) for label in labels):
            raise RuntimeError('Non-max suppression expects "{}" tag in labels.'.format(confidence_tag_name))
        if not all(label.tags.get(confidence_tag_name).meta.value_type == TagValueType.ANY_NUMBER for label in labels):
            raise RuntimeError('Non-max suppression expects "{}" tag to have type {}.'.format(confidence_tag_name,
                                                                                              TagValueType.ANY_NUMBER))

        # Group labels by their respective class.
        labels_per_class = defaultdict(list)
        for label in labels:
            labels_per_class[label.obj_class.name].append(label)
        survived_labels = [cls._single_class_nms(single_class_labels, iou_thresh, confidence_tag_name) for
                           single_class_labels in labels_per_class.values()]
        return list(itertools.chain(*survived_labels))

    def _do_infer_annotate(self, img: np.ndarray, ann: Annotation) -> Annotation:
        result_ann = ann.clone()
        model_labels = []
        roi_bbox_labels = []
        for roi in self._sliding_windows.get(ann.img_size):
            raw_roi_ann = _get_annotation_for_bbox(img, roi, self._model)
            # Accumulate all the labels across the sliding windows to potentially run non-max suppression over them.
            # Only retain the classes that will be eventually saved to avoid running NMS on objects we will
            # throw away anyway.
            model_labels.extend([
                label for label in raw_roi_ann.labels
                if isinstance(label.geometry, Rectangle) and self._model_class_mapper.map(label.obj_class) is not None])

            model_img_level_tags = make_renamed_tags(
                raw_roi_ann.img_tags, self._model_tag_meta_mapper, skip_missing=True)
            roi_bbox_labels.extend(
                _maybe_make_bbox_label(roi, self._intermediate_bbox_class, tags=model_img_level_tags))

        nms_conf = self._config.get(NMS_AFTER, {ENABLE: False})
        if nms_conf[ENABLE]:
            confidence_tag_name = nms_conf.get(CONFIDENCE_TAG_NAME, CONFIDENCE)
            model_labels = self._general_nms(
                labels=model_labels, iou_thresh=nms_conf[IOU_THRESHOLD], confidence_tag_name=confidence_tag_name)

        model_labels_renamed = _replace_or_drop_labels_classes(
            model_labels, self._model_class_mapper, self._model_tag_meta_mapper)

        result_ann = result_ann.add_labels(roi_bbox_labels + model_labels_renamed)
        return result_ann


class InferenceModeFactory:
    mapping = {inference_mode_cls.mode_name(): inference_mode_cls
               for inference_mode_cls in [InfModeFullImage,
                                          InfModeRoi,
                                          InfModeBboxes,
                                          InfModeSlidingWindowSegmentation,
                                          InfModeSlidingWindowDetection]}

    @classmethod
    def create(cls, config, *args, **kwargs):
        key = config[NAME]
        feeder_cls = cls.mapping.get(key)
        if feeder_cls is None:
            raise NotImplementedError()
        res = feeder_cls(config, *args, **kwargs)
        return res


def get_effective_inference_mode_config(task_inference_mode_config: dict, default_inference_mode_config: dict) -> dict:
    task_inference_mode_name = task_inference_mode_config.get(NAME, None)
    default_inference_mode_name = default_inference_mode_config.get(NAME, None)
    inference_mode_name = task_inference_mode_name or default_inference_mode_name
    if inference_mode_name is None:
        raise RuntimeError(
            'Inference mode name ({} key) not found in mode config.'.format(NAME))

    inference_mode_cls = InferenceModeFactory.mapping[inference_mode_name]

    # Priorities when filling out inference mode config:
    # - task config (highest)
    # - default config from constructor
    # - default config for the given mode (lowest).
    result_config = inference_mode_cls.make_default_config(model_result_suffix='')
    if default_inference_mode_name == inference_mode_name:
        # Only take custom mode defaults into account if the mode name matches the task config.
        update_recursively(result_config, default_inference_mode_config)
    update_recursively(result_config, task_inference_mode_config)
    return result_config



# coding: utf-8

from supervisely_lib.api.module_api import ApiField
from supervisely_lib.api.entity_annotation.tag_api import TagApi


class PointcloudTagApi(TagApi):
    _entity_id_field = ApiField.ENTITY_ID
    _method_bulk_add = 'point-clouds.tags.bulk.add'




# coding: utf-8
import random
import colorsys


def _validate_color(color):
    """
    Checks input color for compliance with the required format
    :param: color: color (RGB tuple of integers)
    """
    if not isinstance(color, (list, tuple)):
        raise ValueError('Color has to be list, or tuple')
    if len(color) != 3:
        raise ValueError('Color have to contain exactly 3 values: [R, G, B]')
    for channel in color:
        validate_channel_value(channel)


def random_rgb() -> list:
    """
    Generate RGB color with fixed saturation and lightness
    :return: RGB integer values.
    """
    hsl_color = (random.random(), 0.3, 0.8)
    rgb_color = colorsys.hls_to_rgb(*hsl_color)
    return [round(c * 255) for c in rgb_color]


def _normalize_color(color):
    """
    Divide all RGB values by 255.
    :param color: color (RGB tuple of integers)
    """
    return [c / 255. for c in color]


def _color_distance(first_color: list, second_color: list) -> float:
    """
    Calculate distance in HLS color space between Hue components of 2 colors
    :param first_color: first color (RGB tuple of integers)
    :param second_color: second color (RGB tuple of integers)
    :return: Euclidean distance between 'first_color' and 'second_color'
    """
    first_color_hls = colorsys.rgb_to_hls(*_normalize_color(first_color))
    second_color_hls = colorsys.rgb_to_hls(*_normalize_color(second_color))
    hue_distance = min(abs(first_color_hls[0] - second_color_hls[0]),
                       1 - abs(first_color_hls[0] - second_color_hls[0]))
    return hue_distance


def generate_rgb(exist_colors: list) -> list:
    """
    Generate new color which oppositely by exist colors
    :param exist_colors: list of existing colors in RGB format.
    :return: RGB integer values. Example: [80, 255, 0]
    """
    largest_min_distance = 0
    best_color = random_rgb()
    if len(exist_colors) > 0:
        for _ in range(100):
            color = random_rgb()
            current_min_distance = min(_color_distance(color, c) for c in exist_colors)
            if current_min_distance > largest_min_distance:
                largest_min_distance = current_min_distance
                best_color = color
    _validate_color(best_color)
    return best_color


def rgb2hex(color: list) -> str:
    """
    Convert integer color format to HEX string
    :param color: RGB integer values. Example: [80, 255, 0]
    :return: HEX RGB string. Example: "#FF42А4
    """
    _validate_color(color)
    return '#' + ''.join('{:02X}'.format(component) for component in color)


def _hex2color(hex_value: str) -> list:
    """
        Convert HEX RGB string to integer RGB format
        :param hex_value: HEX RGBA string. Example: "#FF02А4
        :return: RGB integer values. Example: [80, 255, 0]
    """
    assert hex_value.startswith('#')
    return [int(hex_value[i:(i + 2)], 16) for i in range(1, len(hex_value), 2)]


def hex2rgb(hex_value: str) -> list:
    """
    Convert HEX RGB string to integer RGB format
    :param hex_value: HEX RGBA string. Example: "#FF02А4
    :return: RGB integer values. Example: [80, 255, 0]
    """
    assert len(hex_value) == 7, "Supported only HEX RGB string format!"
    color = _hex2color(hex_value)
    _validate_color(color)
    return color


def _hex2rgba(hex_value: str) -> list:
    """
    Convert HEX RGBA string to integer RGBA format
    :param hex_value: HEX RGBA string. Example: "#FF02А4CC
    :return: RGBA integer values. Example: [80, 255, 0, 128]
    """
    assert len(hex_value) == 9, "Supported only HEX RGBA string format!"
    return _hex2color(hex_value)


def validate_channel_value(value):
    if 0 <= value <= 255:
        pass
    else:
        raise ValueError('Color channel has to be in range [0; 255]')


# coding: utf-8
import numpy as np

import torch
import torch.nn.functional as torch_functional

from supervisely_lib.imaging import image as sly_image
from supervisely_lib.nn.pytorch.cuda import cuda_variable


def infer_per_pixel_scores_single_image(model, raw_input, out_shape, apply_softmax=True):
    """
    Performs inference with PyTorch model and resize predictions to a given size.

    Args:
        model: PyTorch model inherited from torch.Module class.
        raw_input: PyTorch Tensor
        out_shape: Output size (height, width).
        apply_softmax: Whether to apply softmax function after inference or not.
    Returns:
        Inference resulting numpy array resized to a given size.
    """
    model_input = torch.stack([raw_input], 0)  # add dim #0 (batch size 1)
    model_input = cuda_variable(model_input, volatile=True)

    output = model(model_input)
    if apply_softmax:
        output = torch_functional.softmax(output, dim=1)
    output = output.data.cpu().numpy()[0]  # from batch to 3d

    pred = np.transpose(output, (1, 2, 0))
    return sly_image.resize(pred, out_shape)



# coding: utf-8

import cv2
import numpy as np

from supervisely_lib.geometry.constants import FACES, POINTS, LABELER_LOGIN, UPDATED_AT, CREATED_AT, ID, CLASS_ID
from supervisely_lib.geometry.geometry import Geometry
from supervisely_lib.geometry.point_location import points_to_row_col_list, row_col_list_to_points
from supervisely_lib.geometry.rectangle import Rectangle


class CuboidFace:
    '''
    This is a class for creating and using CuboidFace objects for Cuboid class objects
    '''
    def __init__(self, a, b, c, d):
        '''
        :param a: int
        :param b: int
        :param c: int
        :param d: int
        '''
        self._a = a
        self._b = b
        self._c = c
        self._d = d

    def to_json(self):
        '''
        The function to_json convert CuboidFace class object to list
        :return: list of integers
        '''
        return [self.a, self.b, self.c, self.d]

    @classmethod
    def from_json(cls, data):
        '''
        The function from_json convert CuboidFace from list to CuboidFace class object.
        :param data: list of integers
        :return: CuboidFace class object
        '''
        if len(data) != 4:
            raise ValueError(f'CuboidFace JSON data must have exactly 4 indices, instead got {len(data)!r}.')
        return cls(data[0], data[1], data[2], data[3])

    @property
    def a(self):
        return self._a

    @property
    def b(self):
        return self._b

    @property
    def c(self):
        return self._c

    @property
    def d(self):
        return self._d

    def tolist(self):
        '''
        The function tolist convert CuboidFace class object to list
        :return: list of integers
        '''
        return [self.a, self.b, self.c, self.d]


class Cuboid(Geometry):
    '''
    This is a class for creating and using Cuboid objects for Labels
    '''
    @staticmethod
    def geometry_name():
        return 'cuboid'

    def __init__(self, points, faces, sly_id=None, class_id=None, labeler_login=None, updated_at=None, created_at=None):
        """
        Args:
            points: iterable of PointLocation objects.
            faces: iterable of CuboidFace objects.
        """

        super().__init__(sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

        points = list(points)
        faces = list(faces)

        if len(faces) != 3:
            raise ValueError(f'A cuboid must have exactly 3 faces. Instead got {len(faces)} faces.')

        for face in faces:
            for point_idx in (face.a, face.b, face.c, face.d):
                if point_idx >= len(points) or point_idx < 0:
                    raise ValueError(f'Point index is out of bounds for cuboid face. Got {len(points)!r} points, but '
                                     f'the index is {point_idx!r}.')

        self._points = points
        self._faces = faces

    """
    Implementation of all methods from Geometry
    """

    @property
    def points(self):
        return self._points.copy()

    @property
    def faces(self):
        return self._faces.copy()

    def to_json(self):
        '''
        The function to_json convert Cuboid class object to json format
        :return: Cuboid in json format
        '''
        packed_obj = {
            POINTS: points_to_row_col_list(self._points, flip_row_col_order=True),
            FACES: [face.to_json() for face in self._faces]
        }
        self._add_creation_info(packed_obj)
        return packed_obj

    @classmethod
    def from_json(cls, data):
        '''
        The function from_json convert Cuboid from json format to Cuboid class object. If json format is not correct it generate exception error.
        :param data: input Cuboid in json format
        :return: Cuboid class object
        '''
        for k in [POINTS, FACES]:
            if k not in data:
                raise ValueError(f'Field {k!r} not found in Cuboid JSON data.')

        points = row_col_list_to_points(data[POINTS], flip_row_col_order=True)
        faces = [CuboidFace.from_json(face_json) for face_json in data[FACES]]

        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)
        return cls(points=points, faces=faces,
                   sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

    def crop(self, rect):
        '''
        Crop the current Cuboid with a given rectangle
        :param rect: Rectangle class object
        :return: list with Cuboid class object if rectangle contain all points of Cuboid and empty list otherwise
        '''
        is_all_nodes_inside = all(
            rect.contains_point_location(self._points[p]) for face in self._faces for p in face.tolist())
        return [self] if is_all_nodes_inside else []

    def _transform(self, transform_fn):
        return Cuboid(points=[transform_fn(p) for p in self.points], faces=self.faces)

    def rotate(self, rotator):
        '''
        The function rotate Cuboid with a given rotator(ImageRotator class object contain size of image and angle to rotate)
        :param rotator: ImageRotator class object
        :return: Cuboid class object
        '''
        return self._transform(lambda p: rotator.transform_point(p))

    def resize(self, in_size, out_size):
        '''
        Resize the current Cuboid to match a certain size
        :param in_size: input image size
        :param out_size: output image size
        :return: Cuboid class object
        '''
        return self._transform(lambda p: p.resize(in_size, out_size))

    def scale(self, factor):
        '''
        The function scale change scale of the current Cuboid object with a given factor
        :param factor: float scale parameter
        :return: Cuboid class object
        '''
        return self._transform(lambda p: p.scale(factor))

    def translate(self, drow, dcol):
        '''
        The function translate shifts the Cuboid by a certain number of pixels
        :param drow: horizontal shift
        :param dcol: vertical shift
        :return: Cuboid class object
        '''
        return self._transform(lambda p: p.translate(drow, dcol))

    def fliplr(self, img_size):
        '''
        The function fliplr the current Cuboid object geometry in horizontal
        :param img_size: size of the image
        :return: Cuboid class object
        '''
        return self._transform(lambda p: p.fliplr(img_size))

    def flipud(self, img_size):
        '''
        The function flipud the current Cuboid object geometry in vertical
        :param img_size: size of the image
        :return: Cuboid class object
        '''
        return self._transform(lambda p: p.flipud(img_size))

    def _draw_impl(self, bitmap: np.ndarray, color, thickness=1, config=None):
        bmp_to_draw = np.zeros(bitmap.shape[:2], np.uint8)
        for contour in self._contours_list():
            cv2.fillPoly(bmp_to_draw, pts=[np.array(contour, dtype=np.int32)], color=1)
        bool_mask = bmp_to_draw.astype(bool)
        bitmap[bool_mask] = color

    def _draw_contour_impl(self, bitmap, color, thickness=1, config=None):
        contours_np_list = [np.array(contour, dtype=np.int32) for contour in self._contours_list()]
        cv2.polylines(bitmap, pts=contours_np_list, isClosed=True, color=color, thickness=thickness)

    def _contours_list(self):
        return [points_to_row_col_list([self._points[idx] for idx in face.tolist()], flip_row_col_order=True)
                for face in self._faces]

    def to_bbox(self):
        '''
        The function to_bbox create Rectangle class object from current Cuboid class object
        :return: Rectangle class object
        '''
        points_np = np.array([[self._points[p].row, self._points[p].col]
                              for face in self._faces for p in face.tolist()])
        rows, cols = points_np[:, 0], points_np[:, 1]
        return Rectangle(top=round(min(rows).item()), left=round(min(cols).item()), bottom=round(max(rows).item()),
                         right=round(max(cols).item()))

    @property
    def area(self):
        '''
        :return: area of current Cuboid object
        '''
        bbox = self.to_bbox()
        canvas = np.zeros([bbox.bottom + 1, bbox.right + 1], dtype=np.bool)
        self.draw(canvas, True)
        return float(np.sum(canvas))



# coding: utf-8

import cv2
from copy import deepcopy

from supervisely_lib.imaging.color import rgb2hex, hex2rgb
from supervisely_lib.io.json import JsonSerializable

from supervisely_lib.geometry.point import Point
from supervisely_lib.geometry.point_location import PointLocation
from supervisely_lib.geometry.rectangle import Rectangle
from supervisely_lib.geometry.geometry import Geometry
from supervisely_lib.geometry.constants import LABELER_LOGIN, CREATED_AT, UPDATED_AT, ID, CLASS_ID


EDGES = 'edges'
NODES = 'nodes'

DISABLED = 'disabled'
LOC = 'loc'

DST = 'dst'
SRC = 'src'
COLOR = 'color'


class Node(JsonSerializable):
    '''
    This is a class for creating and using Nodes
    '''
    def __init__(self, location: PointLocation, disabled=True):
        '''
        :param location: PointLocation class object
        :param disabled: bool
        '''
        self._location = location
        self._disabled = disabled

    @property
    def location(self):
        return self._location

    @property
    def disabled(self):
        return self._disabled

    @classmethod
    def from_json(cls, data):
        '''
        The function from_json convert Node from json format to Node class object.
        :param data: input node in json format
        :return: Node class object
        '''
        # TODO validations
        loc = data[LOC]
        return cls(location=PointLocation(row=loc[1], col=loc[0]), disabled=data.get(DISABLED, False))

    def to_json(self):
        '''
        The function to_json convert node to json format
        :return: node in json format
        '''
        result = {LOC: [self._location.col, self._location.row]}
        if self.disabled:
            result[DISABLED] = True
        return result

    def transform_location(self, transform_fn):
        '''
        :param transform_fn: function to convert location
        :return: Node class object with the changed location attribute using the given function
        '''
        return Node(transform_fn(self._location), disabled=self.disabled)


def _maybe_transform_colors(elements, process_fn):
    '''
    Function _maybe_transform_colors convert some list of parameters using the given function
    :param elements: list of elements
    :param process_fn: function to convert
    '''
    for elem in elements:
        if COLOR in elem:
            elem[COLOR] = process_fn(elem[COLOR])


class GraphNodes(Geometry):
    '''
    This is a class for creating and using GraphNodes
    '''
    @staticmethod
    def geometry_name():
        return 'graph'

    def __init__(self, nodes: dict,
                 sly_id=None, class_id=None, labeler_login=None, updated_at=None, created_at=None):
        '''
        :param nodes: dictionary containing nodes of graph
        '''
        super().__init__(sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)
        self._nodes = nodes

    @property
    def nodes(self):
        return self._nodes.copy()

    @classmethod
    def from_json(cls, data):
        '''
        The function from_json convert GraphNodes from json format to GraphNodes class object.
        :param data: input graph in json format
        :return: GraphNodes class object
        '''
        nodes = {node_id: Node.from_json(node_json) for node_id, node_json in data['nodes'].items()}
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, Node)
        return GraphNodes(nodes=nodes, sly_id=sly_id, class_id=class_id,
                          labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

    def to_json(self):
        '''
        The function to_json convert graph to json format
        :return: graph in json format
        '''
        res = {NODES: {node_id: node.to_json() for node_id, node in self._nodes.items()}}
        self._add_creation_info(res)
        return res

    def crop(self, rect: Rectangle):
        '''
        The function "crop" return list containing graph if all nodes of graph located in given rectangle and an empty list otherwise
        :param rect: Rectangle class object
        :return: list containing GraphNodes class object or empty list
        '''
        is_all_nodes_inside = all(rect.contains_point_location(node.location) for node in self._nodes.values())
        return [self] if is_all_nodes_inside else []

    def relative_crop(self, rect):
        '''
        The function relative_crop calculates new parameters of graph nodes after shifts it with given rectangle(on value of it left top angle)
        :param rect: Rectangle class object
        :return: GraphNodes class object
        '''
        return [geom.translate(drow=-rect.top, dcol=-rect.left) for geom in self.crop(rect)]

    def transform(self, transform_fn):
        '''
        The function "transform" transform graph nodes with given function
        :param transform_fn: function to convert
        :return: GraphNodes class object
        '''
        return GraphNodes(nodes={node_id: transform_fn(node) for node_id, node in self._nodes.items()})

    def transform_locations(self, transform_fn):
        '''
        :param transform_fn: function to convert location
        :return: GraphNodes class object with the changed location attribute of the nodes using the given function
        '''
        return self.transform(lambda kp: kp.transform_location(transform_fn))

    def resize(self, in_size, out_size):
        '''
        The function resize calculates new values of graph nodes after resizing graph
        :param in_size: old image size
        :param out_size: new image size
        :return: GraphNodes class object
        '''
        return self.transform_locations(lambda p: p.resize(in_size, out_size))

    def scale(self, factor):
        '''
        The function scale calculates new values of graph nodes after scaling graph with given factor
        :param factor: float scale parameter
        :return: GraphNodes class object
        '''
        return self.transform_locations(lambda p: p.scale(factor))

    def translate(self, drow, dcol):
        '''
        The function translate calculates new values of graph nodes after shifts it by a certain number of pixels
        :param drow: int
        :param dcol: int
        :return: GraphNodes class object
        '''
        return self.transform_locations(lambda p: p.translate(drow, dcol))

    def rotate(self, rotator):
        '''
        The function rotate calculates new values of graph nodes after rotating graph
        :param rotator: ImageRotator class object
        :return: GraphNodes class object
        '''
        return self.transform_locations(lambda p: p.rotate(rotator))

    def fliplr(self, img_size):
        '''
        The function fliplr calculates new values of graph nodes after fliping graph in horizontal
        :param img_size: tuple or list of integers (image size)
        :return: GraphNodes class object
        '''
        return self.transform_locations(lambda p: p.fliplr(img_size))

    def flipud(self, img_size):
        '''
        The function fliplr calculates new values of graph nodes after fliping graph in vertical
        :param img_size: tuple or list of integers (image size)
        :return: GraphNodes class object
        '''
        return self.transform_locations(lambda p: p.flipud(img_size))

    def _draw_impl(self, bitmap, color, thickness=1, config=None):
        '''
        Draws the graph contour on a given bitmap canvas
        :param bitmap: numpy array
        :param color: tuple or list of integers
        :param thickness: int
        :param config: drawing config specific to a concrete subclass, e.g. per edge colors
        '''
        self.draw_contour(bitmap, color, thickness, config=config)

    @staticmethod
    def _get_nested_or_default(dict, keys_path, default=None):
        result = dict
        for key in keys_path:
            if result is not None:
                result = result.get(key, None)
        return result if result is not None else default

    def _draw_contour_impl(self, bitmap, color=None, thickness=1, config=None):
        if config is not None:
            # If a config with edges and colors is passed, make sure it is
            # consistent with the our set of points.
            self.validate(self.geometry_name(), config)

        # Draw edges first so that nodeas are then drawn on top.
        for edge in self._get_nested_or_default(config, [EDGES], []):
            src = self._nodes.get(edge[SRC], None)
            dst = self._nodes.get(edge[DST], None)
            if (src is not None) and (not src.disabled) and (dst is not None) and (not dst.disabled):
                edge_color = edge.get(COLOR, color)
                cv2.line(bitmap,
                         (src.location.col, src.location.row),
                         (dst.location.col, dst.location.row),
                         tuple(edge_color),
                         thickness)

        nodes_config = self._get_nested_or_default(config, [NODES])
        for node_id, node in self._nodes.items():
            if not node.disabled:
                effective_color = self._get_nested_or_default(nodes_config, [node_id, COLOR], color)
                Point.from_point_location(node.location).draw(
                    bitmap=bitmap, color=effective_color, thickness=thickness, config=None)

    @property
    def area(self):
        return 0.0

    def to_bbox(self):
        '''
        The function to_bbox create Rectangle class object from current GraphNodes class object
        :return: Rectangle class object
        '''
        return Rectangle.from_geometries_list(
            [Point.from_point_location(node.location) for node in self._nodes.values()])

    def clone(self):
        return self

    def validate(self, name, settings):
        '''
        Checks the graph for correctness and compliance with the template
        '''
        super().validate(name, settings)
        # TODO template self-consistency checks.

        nodes_not_in_template = set(self._nodes.keys()) - set(settings[NODES].keys())
        if len(nodes_not_in_template) > 0:
            raise ValueError('Graph contains nodes not declared in the template: {!r}.'.format(nodes_not_in_template))

    @staticmethod
    def _transform_config_colors(config, transform_fn):
        '''
        Transform colors of edges and nodes in graph template
        :param config: dictionary(graph template)
        :param transform_fn: function to convert
        :return: dictionary(graph template)
        '''
        if config is None:
            return None

        result = deepcopy(config)
        _maybe_transform_colors(result.get(EDGES, []), transform_fn)
        _maybe_transform_colors(result[NODES].values(), transform_fn)
        return result

    @staticmethod
    def config_from_json(config):
        '''
        Convert graph template from json format
        :param config: dictionary(graph template) in json format
        :return: dictionary(graph template)
        '''
        return GraphNodes._transform_config_colors(config, hex2rgb)

    @staticmethod
    def config_to_json(config):
        '''
        Convert graph template in json format
        :param config: dictionary(graph template)
        :return: dictionary(graph template) in json format
        '''
        return GraphNodes._transform_config_colors(config, rgb2hex)

    @classmethod
    def allowed_transforms(cls):
        from supervisely_lib.geometry.any_geometry import AnyGeometry
        from supervisely_lib.geometry.rectangle import Rectangle
        return [AnyGeometry, Rectangle]



# coding: utf-8
from supervisely_lib.api.module_api import ModuleApiBase, ApiField
from supervisely_lib.collection.str_enum import StrEnum
from supervisely_lib.io.fs import ensure_base_path


class Provider(StrEnum):
    S3 = 's3'
    GOOGLE = 'google'
    AZURE = 'azure'

    @staticmethod
    def validate_path(path):
        if not path.startswith(str(Provider.S3)) and \
            not path.startswith(str(Provider.GOOGLE)) and \
            not path.startswith(str(Provider.AZURE)): \
            raise ValueError("Incorrect cloud path, learn more here: https://docs.supervise.ly/enterprise-edition/advanced-tuning/s3#links-plugin-cloud-providers-support")


class RemoteStorageApi(ModuleApiBase):
    def _convert_json_info(self, info: dict):
        return info

    def list(self, path, recursive=True, files=True, folders=True):
        Provider.validate_path(path)
        resp = self._api.get('remote-storage.list', {
            ApiField.PATH: path,
            'recursive': recursive,
            'files': files,
            'folders': folders
        })
        return resp.json()

    def download_path(self, remote_path, save_path, progress_cb=None):
        ensure_base_path(save_path)
        response = self._api.post('remote-storage.download', {ApiField.LINK: remote_path}, stream=True)
        # if "Content-Length" in response.headers:
        #     length = int(response.headers['Content-Length'])
        with open(save_path, 'wb') as fd:
            for chunk in response.iter_content(chunk_size=1024*1024):
                fd.write(chunk)
                if progress_cb is not None:
                    progress_cb(len(chunk))



# coding: utf-8

from collections import namedtuple
import os
import json
from enum import Enum

from supervisely_lib.annotation.annotation import Annotation, ANN_EXT
from supervisely_lib.collection.key_indexed_collection import KeyIndexedCollection, KeyObject
from supervisely_lib.imaging import image as sly_image
from supervisely_lib.io.fs import list_files, list_files_recursively, mkdir, copy_file, get_subdirs, dir_exists
from supervisely_lib.io.json import dump_json_file, load_json_file
from supervisely_lib.project.project_meta import ProjectMeta
from supervisely_lib.task.progress import Progress
from supervisely_lib._utils import batched
from supervisely_lib.io.fs import file_exists
from supervisely_lib.api.api import Api
from supervisely_lib.sly_logger import logger
from supervisely_lib.io.fs_cache import FileCache


# @TODO: rename img_path to item_path
ItemPaths = namedtuple('ItemPaths', ['img_path', 'ann_path'])


class OpenMode(Enum):
    READ = 1
    CREATE = 2


def _get_effective_ann_name(img_name, ann_names):
    new_format_name = img_name + ANN_EXT
    if new_format_name in ann_names:
        return new_format_name
    else:
        old_format_name = os.path.splitext(img_name)[0] + ANN_EXT
        return old_format_name if (old_format_name in ann_names) else None


class Dataset(KeyObject):
    '''
    This is a class for creating and using Dataset objects. Here is where your labeled and unlabeled images and other
    files live. There is no more levels: images or videos are directly attached to a dataset. Dataset is a unit of work.
    All images or videos are directly attached to a dataset. A dataset is some sort of data folder with stuff to annotate.
    '''
    item_dir_name = 'img'
    annotation_class = Annotation

    def __init__(self, directory: str, mode: OpenMode):
        '''
        :param directory: path to the directory where the data set will be saved or where it will be loaded from
        :param mode: OpenMode class object which determines in what mode to work with the dataset
        '''
        if type(mode) is not OpenMode:
            raise TypeError("Argument \'mode\' has type {!r}. Correct type is OpenMode".format(type(mode)))

        self._directory = directory
        self._item_to_ann = {} # item file name -> annotation file name

        project_dir, ds_name = os.path.split(directory.rstrip('/'))
        self._project_dir = project_dir
        self._name = ds_name

        if mode is OpenMode.READ:
            self._read()
        else:
            self._create()

    @property
    def name(self):
        return self._name

    def key(self):
        return self.name

    @property
    def directory(self):
        return self._directory

    @property
    def item_dir(self):
        return self.img_dir

    @property
    def img_dir(self):
        '''
        :return: path to the directory with images in dataset
        '''
        # @TODO: deprecated method, should be private and be renamed in future
        return os.path.join(self.directory, self.item_dir_name)

    @property
    def ann_dir(self):
        '''
        :return: path to the directory with annotations in dataset
        '''
        return os.path.join(self.directory, 'ann')

    @staticmethod
    def _has_valid_ext(path: str) -> bool:
        '''
        The function _has_valid_ext checks if a given file has a supported extension('.jpg', '.jpeg', '.mpo', '.bmp', '.png', '.webp')
        :param path: the path to the file
        :return: bool (True if a given file has a supported extension, False - in otherwise)
        '''
        return sly_image.has_valid_ext(path)

    def _read(self):
        '''
        Fills out the dictionary items: item file name -> annotation file name. Checks item and annotation directoris existing and dataset not empty.
        Consistency checks. Every image must have an annotation, and the correspondence must be one to one.
        If not - it generate exception error.
        '''
        if not dir_exists(self.item_dir):
            raise FileNotFoundError('Item directory not found: {!r}'.format(self.item_dir))
        if not dir_exists(self.ann_dir):
            raise FileNotFoundError('Annotation directory not found: {!r}'.format(self.ann_dir))

        raw_ann_paths = list_files(self.ann_dir, [ANN_EXT])
        img_paths = list_files(self.item_dir, filter_fn=self._has_valid_ext)

        raw_ann_names = set(os.path.basename(path) for path in raw_ann_paths)
        img_names = [os.path.basename(path) for path in img_paths]

        if len(img_names) == 0 or len(raw_ann_names) == 0:
            raise RuntimeError('Dataset {!r} is empty'.format(self.name))

        # Consistency checks. Every image must have an annotation, and the correspondence must be one to one.
        effective_ann_names = set()
        for img_name in img_names:
            ann_name = _get_effective_ann_name(img_name, raw_ann_names)
            if ann_name is None:
                raise RuntimeError('Item {!r} in dataset {!r} does not have a corresponding annotation file.'.format(
                    img_name, self.name))
            if ann_name in effective_ann_names:
                raise RuntimeError('Annotation file {!r} in dataset {!r} matches two different image files.'.format(
                    ann_name, self.name))
            effective_ann_names.add(ann_name)
            self._item_to_ann[img_name] = ann_name

    def _create(self):
        '''
        Creates a leaf directory and all intermediate ones for items and annatations.
        '''
        mkdir(self.ann_dir)
        mkdir(self.item_dir)

    def item_exists(self, item_name):
        '''
        Checks if given item name belongs to items of dataset
        :param item_name: str
        :return: bool
        '''
        return item_name in self._item_to_ann

    def get_item_path(self, item_name):
        '''
        :param item_name: str
        :return: str (path to given item), generate exception error if item not found in dataset
        '''
        return self.get_img_path(item_name)

    def get_img_path(self, item_name):
        '''
        :param item_name: str
        :return: str (path to given image), generate exception error if image not found in dataset
        '''
        # @TODO: deprecated method, should be private and be renamed in future
        if not self.item_exists(item_name):
            raise RuntimeError('Item {} not found in the project.'.format(item_name))
        return os.path.join(self.item_dir, item_name)

    def get_ann_path(self, item_name):
        '''
        :param item_name: str
        :return: str (path to given annotation), generate exception error if annotation not found in dataset
        '''
        ann_path = self._item_to_ann.get(item_name, None)
        if ann_path is None:
            raise RuntimeError('Item {} not found in the project.'.format(item_name))
        return os.path.join(self.ann_dir, ann_path)

    def add_item_file(self, item_name, item_path, ann=None, _validate_item=True, _use_hardlink=False):
        '''
        Add given item file to dataset items directory, and add given annatation to dataset annotations dir(if ann=None
        empty annotation will be create). Generate exception error if item_name already exists in dataset or item name has unsupported extension
        :param item_name: str
        :param item_path: str
        :param ann: Annotation class object, str, dict, None (generate exception error if param type is another)
        :param _validate_item: bool
        :param _use_hardlink: bool
        '''
        self._add_item_file(item_name, item_path, _validate_item=_validate_item, _use_hardlink=_use_hardlink)
        self._add_ann_by_type(item_name, ann)

    def add_item_np(self, item_name, img, ann=None):
        '''
        Write given image(RGB format(numpy matrix)) to dataset items directory, and add given annatation to dataset annotations dir(if ann=None
        empty annotation will be create). Generate exception error if item_name already exists in dataset or item name has unsupported extension
        :param item_name: str
        :param img: image in RGB format(numpy matrix)
        :param ann: Annotation class object, str, dict, None (generate exception error if param type is another)
        '''
        self._add_img_np(item_name, img)
        self._add_ann_by_type(item_name, ann)

    def add_item_raw_bytes(self, item_name, item_raw_bytes, ann=None):
        '''
        Write given binary object to dataset items directory, and add given annatation to dataset annotations dir(if ann=None
        empty annotation will be create). Generate exception error if item_name already exists in dataset or item name has unsupported extension.
        Make sure we actually received a valid image file, clean it up and fail if not so.
        :param item_name: str
        :param item_raw_bytes: binary object
        :param ann: Annotation class object, str, dict, None (generate exception error if param type is another)
        '''
        self._add_item_raw_bytes(item_name, item_raw_bytes)
        self._add_ann_by_type(item_name, ann)

    def _get_empty_annotaion(self, item_name):
        '''
        Create empty annotation from given item. Generate exception error if item not found in project
        :param item_name: str
        :return: Annotation class object
        '''
        img_size = sly_image.read(self.get_img_path(item_name)).shape[:2]
        return self.annotation_class(img_size)

    def _add_ann_by_type(self, item_name, ann):
        '''
        Add given annatation to dataset annotations dir and to dictionary items: item file name -> annotation file name
        :param item_name: str
        :param ann: Annotation class object, str, dict, None (generate exception error if param type is another)
        '''
        # This is a new-style annotation name, so if there was no image with this name yet, there should not have been
        # an annotation either.
        self._item_to_ann[item_name] = item_name + ANN_EXT
        if ann is None:
            self.set_ann(item_name, self._get_empty_annotaion(item_name))
        elif type(ann) is self.annotation_class:
            self.set_ann(item_name, ann)
        elif type(ann) is str:
            self.set_ann_file(item_name, ann)
        elif type(ann) is dict:
            self.set_ann_dict(item_name, ann)
        else:
            raise TypeError("Unsupported type {!r} for ann argument".format(type(ann)))

    def _check_add_item_name(self, item_name):
        '''
        Generate exception error if item name already exists in dataset or has unsupported extension
        :param item_name: str
        '''
        if item_name in self._item_to_ann:
            raise RuntimeError('Item {!r} already exists in dataset {!r}.'.format(item_name, self.name))
        if not self._has_valid_ext(item_name):
            raise RuntimeError('Item name {!r} has unsupported extension.'.format(item_name))

    def _add_item_raw_bytes(self, item_name, item_raw_bytes):
        '''
        Write given binary object to dataset items directory, Generate exception error if item_name already exists in
        dataset or item name has unsupported extension. Make sure we actually received a valid image file, clean it up and fail if not so.
        :param item_name: str
        :param item_raw_bytes: binary object
        '''
        self._check_add_item_name(item_name)
        dst_img_path = os.path.join(self.item_dir, item_name)
        with open(dst_img_path, 'wb') as fout:
            fout.write(item_raw_bytes)
        self._validate_added_item_or_die(dst_img_path)

    def generate_item_path(self, item_name):
        '''
        :param item_name: str
        :return: str (full path to the given item)
        '''
        return os.path.join(self.item_dir, item_name)

    def _add_img_np(self, item_name, img):
        '''
        Write given image(RGB format(numpy matrix)) to dataset items directory. Generate exception error if item_name
        already exists in dataset or item name has unsupported extension
        :param item_name: str
        :param img: image in RGB format(numpy matrix)
        '''
        self._check_add_item_name(item_name)
        dst_img_path = os.path.join(self.item_dir, item_name)
        sly_image.write(dst_img_path, img)

    def _add_item_file(self, item_name, item_path, _validate_item=True, _use_hardlink=False):
        self._add_img_file(item_name, item_path, _validate_item, _use_hardlink)

    def _add_img_file(self, item_name, img_path, _validate_img=True, _use_hardlink=False):
        '''
        Add given item file to dataset items directory. Generate exception error if item_name already exists in dataset
        or item name has unsupported extension
        :param item_name: str
        :param img_path: str
        :param _validate_img: bool
        :param _use_hardlink: bool
        '''
        # @TODO: deprecated method, should be private and be (refactored, renamed) in future
        self._check_add_item_name(item_name)
        dst_img_path = os.path.join(self.item_dir, item_name)
        if img_path != dst_img_path and img_path is not None:  # used only for agent + api during download project + None to optimize internal usage
            hardlink_done = False
            if _use_hardlink:
                try:
                    os.link(img_path, dst_img_path)
                    hardlink_done = True
                except OSError:
                    pass
            if not hardlink_done:
                copy_file(img_path, dst_img_path)
            if _validate_img:
                self._validate_added_item_or_die(img_path)

    @staticmethod
    def _validate_added_item_or_die(item_path):
        '''
        Make sure we actually received a valid image file, clean it up and fail if not so
        :param item_path: str
        '''
        # Make sure we actually received a valid image file, clean it up and fail if not so.
        try:
            sly_image.validate_format(item_path)
        except (sly_image.UnsupportedImageFormat, sly_image.ImageReadException):
            os.remove(item_path)
            raise

    def set_ann(self, item_name: str, ann):
        '''
        Save given annotation with given name to dataset annotations dir in json format.
        :param item_name: str
        :param ann: Annotation class object (Generate exception error if not so)
        '''
        if type(ann) is not self.annotation_class:
            raise TypeError("Type of 'ann' have to be Annotation, not a {}".format(type(ann)))
        dst_ann_path = self.get_ann_path(item_name)
        dump_json_file(ann.to_json(), dst_ann_path, indent=4)

    def set_ann_file(self, item_name: str, ann_path: str):
        '''
        Copy given annotation with given name to dataset annotations dir
        :param item_name: str
        :param ann_path: str (Generate exception error if not so)
        '''
        if type(ann_path) is not str:
            raise TypeError("Annotation path should be a string, not a {}".format(type(ann_path)))
        dst_ann_path = self.get_ann_path(item_name)
        copy_file(ann_path, dst_ann_path)

    def set_ann_dict(self, item_name: str, ann: dict):
        '''
        Save given annotation with given name to dataset annotations dir in json format.
        :param item_name: str
        :param ann: dict (json format)
        '''
        if type(ann) is not dict:
            raise TypeError("Ann should be a dict, not a {}".format(type(ann)))
        dst_ann_path = self.get_ann_path(item_name)
        dump_json_file(ann, dst_ann_path, indent=4)

    def get_item_paths(self, item_name) -> ItemPaths:
        '''
        Create ItemPaths object with passes to items and annotation dirs with given name
        :param item_name: str
        :return: ItemPaths class object
        '''
        return ItemPaths(img_path=self.get_img_path(item_name), ann_path=self.get_ann_path(item_name))

    def __len__(self):
        return len(self._item_to_ann)

    def __next__(self):
        for item_name in self._item_to_ann.keys():
            yield item_name

    def __iter__(self):
        return next(self)


class Project:
    '''
    This is a class for creating and using Project objects. You can think of a Project as a superfolder with data and
    meta information.
    '''
    dataset_class = Dataset
    class DatasetDict(KeyIndexedCollection):
        item_type = Dataset

    def __init__(self, directory, mode: OpenMode):
        '''
        :param directory: path to the directory where the project will be saved or where it will be loaded from
        :param mode: OpenMode class object which determines in what mode to work with the project (generate exception error if not so)
        '''
        if type(mode) is not OpenMode:
            raise TypeError("Argument \'mode\' has type {!r}. Correct type is OpenMode".format(type(mode)))

        parent_dir, name = Project._parse_path(directory)
        self._parent_dir = parent_dir
        self._name = name
        self._datasets = Project.DatasetDict()  # ds_name -> dataset object
        self._meta = None

        if mode is OpenMode.READ:
            self._read()
        else:
            self._create()

    @property
    def parent_dir(self):
        return self._parent_dir

    @property
    def name(self):
        return self._name

    @property
    def datasets(self):
        return self._datasets

    @property
    def meta(self):
        return self._meta

    @property
    def directory(self):
        return os.path.join(self.parent_dir, self.name)

    @property
    def total_items(self):
        '''
        :return: total number of items in project
        '''
        return sum(len(ds) for ds in self._datasets)

    def _get_project_meta_path(self):
        '''
        :return: str (path to project meta file(meta.json))
        '''
        return os.path.join(self.directory, 'meta.json')

    def _read(self):
        '''
        Download project from given project directory. Checks item and annotation directoris existing and dataset not empty.
        Consistency checks. Every image must have an annotation, and the correspondence must be one to one.
        '''
        meta_json = load_json_file(self._get_project_meta_path())
        self._meta = ProjectMeta.from_json(meta_json)

        possible_datasets = get_subdirs(self.directory)
        for ds_name in possible_datasets:
            current_dataset = self.dataset_class(os.path.join(self.directory, ds_name), OpenMode.READ)
            self._datasets = self._datasets.add(current_dataset)

        if self.total_items == 0:
            raise RuntimeError('Project is empty')

    def _create(self):
        '''
        Creates a leaf directory and empty meta.json file. Generate exception error if project directory already exists and is not empty.
        '''
        if dir_exists(self.directory):
            if len(list_files_recursively(self.directory)) > 0:
                raise RuntimeError(
                    "Cannot create new project {!r}. Directory {!r} already exists and is not empty".format(
                        self.name, self.directory))
        else:
            mkdir(self.directory)
        self.set_meta(ProjectMeta())

    def validate(self):
        # @TODO: validation here
        pass

    def set_meta(self, new_meta):
        '''
        Save given meta to project dir in json format.
        :param new_meta: ProjectMeta class object
        '''
        self._meta = new_meta
        dump_json_file(self.meta.to_json(), self._get_project_meta_path(), indent=4)

    def __iter__(self):
        return next(self)

    def __next__(self):
        for dataset in self._datasets:
            yield dataset

    def create_dataset(self, ds_name):
        '''
        Creates a leaf directory with given name and all intermediate ones for items and annatations. Add new dataset
        to the collection of all datasets in project
        :param ds_name: str
        :return: Dataset class object
        '''
        ds = self.dataset_class(os.path.join(self.directory, ds_name), OpenMode.CREATE)
        self._datasets = self._datasets.add(ds)
        return ds

    def _add_item_file_to_dataset(self, ds, item_name, item_paths, _validate_item, _use_hardlink):
        '''
        Add item file and annotation from given name and path to given dataset items directory. Generate exception error if item_name already exists in dataset or item name has unsupported extension
        :param ds: Dataset class object
        :param item_name: str
        :param item_paths: ItemPaths object
        :param _validate_item: bool
        :param _use_hardlink: bool
        '''
        ds.add_item_file(item_name, item_paths.img_path,
                         ann=item_paths.ann_path, _validate_item=_validate_item, _use_hardlink=_use_hardlink)

    def copy_data(self, dst_directory, dst_name=None, _validate_item=True, _use_hardlink=False):
        '''
        Make copy of project in given directory.
        :param dst_directory: str
        :param dst_name: str
        :param _validate_item: bool
        :param _use_hardlink: bool
        :return: Project class object
        '''
        dst_name = dst_name if dst_name is not None else self.name
        new_project = Project(os.path.join(dst_directory, dst_name), OpenMode.CREATE)
        new_project.set_meta(self.meta)

        for ds in self:
            new_ds = new_project.create_dataset(ds.name)
            for item_name in ds:
                item_paths = ds.get_item_paths(item_name)
                self._add_item_file_to_dataset(new_ds, item_name, item_paths, _validate_item, _use_hardlink)
        return new_project

    @staticmethod
    def _parse_path(project_dir):
        '''
        Split given path to project on parent directory and directory where project is located
        :param project_dir: str
        :return: str, str
        '''
        #alternative implementation
        #temp_parent_dir = os.path.dirname(parent_dir)
        #temp_name = os.path.basename(parent_dir)

        parent_dir, pr_name = os.path.split(project_dir.rstrip('/'))
        if not pr_name:
            raise RuntimeError('Unable to determine project name.')
        return parent_dir, pr_name


def read_single_project(dir, project_class=Project):
    '''
    Read project from given ditectory. Generate exception error if given dir contains more than one subdirectory
    :param dir: str
    :param project_class: Project class object type
    :return: Project class object
    '''
    projects_in_dir = get_subdirs(dir)
    if len(projects_in_dir) != 1:
        raise RuntimeError('Found {} dirs instead of 1'.format(len(projects_in_dir)))

    project_dir = os.path.join(dir, projects_in_dir[0])
    try:
        project_fs = project_class(project_dir, OpenMode.READ)
    except Exception as e:
        projects_in_dir = get_subdirs(project_dir)
        if len(projects_in_dir) != 1:
            raise e
        project_dir = os.path.join(project_dir, projects_in_dir[0])
        project_fs = project_class(project_dir, OpenMode.READ)

    return project_fs


def _download_project(api, project_id, dest_dir, dataset_ids=None, log_progress=False, batch_size=10):
    dataset_ids = set(dataset_ids) if (dataset_ids is not None) else None
    project_fs = Project(dest_dir, OpenMode.CREATE)
    meta = ProjectMeta.from_json(api.project.get_meta(project_id))
    project_fs.set_meta(meta)

    for dataset_info in api.dataset.get_list(project_id):
        dataset_id = dataset_info.id
        if dataset_ids is not None and dataset_id not in dataset_ids:
            continue

        dataset_fs = project_fs.create_dataset(dataset_info.name)
        images = api.image.get_list(dataset_id)

        ds_progress = None
        if log_progress:
            ds_progress = Progress(
                'Downloading dataset: {!r}'.format(dataset_info.name), total_cnt=len(images))

        for batch in batched(images, batch_size):
            image_ids = [image_info.id for image_info in batch]
            image_names = [image_info.name for image_info in batch]

            # download images in numpy format
            batch_imgs_bytes = api.image.download_bytes(dataset_id, image_ids)

            # download annotations in json format
            ann_infos = api.annotation.download_batch(dataset_id, image_ids)
            ann_jsons = [ann_info.annotation for ann_info in ann_infos]

            for name, img_bytes, ann in zip(image_names, batch_imgs_bytes, ann_jsons):
                dataset_fs.add_item_raw_bytes(name, img_bytes, ann)

            if log_progress:
                ds_progress.iters_done_report(len(batch))


def upload_project(dir, api, workspace_id, project_name=None, log_progress=True):
    project_fs = read_single_project(dir)
    if project_name is None:
        project_name = project_fs.name

    if api.project.exists(workspace_id, project_name):
        project_name = api.project.get_free_name(workspace_id, project_name)

    project = api.project.create(workspace_id, project_name, )
    api.project.update_meta(project.id, project_fs.meta.to_json())

    for dataset_fs in project_fs.datasets:
        dataset = api.dataset.create(project.id, dataset_fs.name)

        names, img_paths, ann_paths = [], [], []
        for item_name in dataset_fs:
            img_path, ann_path = dataset_fs.get_item_paths(item_name)
            names.append(item_name)
            img_paths.append(img_path)
            ann_paths.append(ann_path)

        progress_cb = None
        if log_progress:
            ds_progress = Progress('Uploading images to dataset {!r}'.format(dataset.name), total_cnt=len(img_paths))
            progress_cb = ds_progress.iters_done_report
        img_infos = api.image.upload_paths(dataset.id, names, img_paths, progress_cb)
        image_ids = [img_info.id for img_info in img_infos]

        if log_progress:
            ds_progress = Progress('Uploading annotations to dataset {!r}'.format(dataset.name), total_cnt=len(img_paths))
            progress_cb = ds_progress.iters_done_report
        api.annotation.upload_paths(image_ids, ann_paths, progress_cb)

    return project.id, project.name


def download_project(api: Api, project_id, dest_dir, dataset_ids=None, log_progress=False, batch_size=10,
                     cache: FileCache = None, progress_cb=None):
    if cache is None:
        _download_project(api, project_id, dest_dir, dataset_ids, log_progress, batch_size)
    else:
        _download_project_optimized(api, project_id, dest_dir, dataset_ids, cache, progress_cb)


def _download_project_optimized(api: Api, project_id, project_dir, datasets_whitelist=None, cache=None, progress_cb=None):
    project_info = api.project.get_info_by_id(project_id)
    project_id = project_info.id
    logger.info(f"Annotations are not cached (always download latest version from server)")
    project_fs = Project(project_dir, OpenMode.CREATE)
    meta = ProjectMeta.from_json(api.project.get_meta(project_id))
    project_fs.set_meta(meta)
    for dataset_info in api.dataset.get_list(project_id):
        dataset_name = dataset_info.name
        dataset_id = dataset_info.id
        need_download = True
        if datasets_whitelist is not None and dataset_id not in datasets_whitelist:
            need_download = False
        if need_download is True:
            dataset = project_fs.create_dataset(dataset_name)
            _download_dataset(api, dataset, dataset_id, cache=cache, progress_cb=progress_cb)


def _split_images_by_cache(images, cache):
    images_to_download = []
    images_in_cache = []
    images_cache_paths = []
    for image in images:
        _, effective_ext = os.path.splitext(image.name)
        if len(effective_ext) == 0:
            # Fallback for the old format where we were cutting off extensions from image names.
            effective_ext = image.ext
        cache_path = cache.check_storage_object(image.hash, effective_ext)
        if cache_path is None:
            images_to_download.append(image)
        else:
            images_in_cache.append(image)
            images_cache_paths.append(cache_path)
    return images_to_download, images_in_cache, images_cache_paths


def _maybe_append_image_extension(name, ext):
    name_split = os.path.splitext(name)
    if name_split[1] == '':
        normalized_ext = ('.' + ext).replace('..', '.')
        result = name + normalized_ext
        sly_image.validate_ext(result)
    else:
        result = name
    return result


def _download_dataset(api: Api, dataset, dataset_id, cache=None, progress_cb=None):
    images = api.image.get_list(dataset_id)

    images_to_download = images

    # copy images from cache to task folder and download corresponding annotations
    if cache:
        images_to_download, images_in_cache, images_cache_paths = _split_images_by_cache(images, cache)
        if len(images_to_download) + len(images_in_cache) != len(images):
            raise RuntimeError("Error with images cache during download. Please contact support.")
        logger.info(f"Download dataset: {dataset.name}", extra={"total": len(images),
                                                                "in cache": len(images_in_cache),
                                                                "to download": len(images_to_download)})
        if len(images_in_cache) > 0:
            img_cache_ids = [img_info.id for img_info in images_in_cache]
            ann_info_list = api.annotation.download_batch(dataset_id, img_cache_ids, progress_cb)
            img_name_to_ann = {ann.image_id: ann.annotation for ann in ann_info_list}
            for batch in batched(list(zip(images_in_cache, images_cache_paths)), batch_size=50):
                for img_info, img_cache_path in batch:
                    item_name = _maybe_append_image_extension(img_info.name, img_info.ext)
                    dataset.add_item_file(item_name, img_cache_path, img_name_to_ann[img_info.id], _validate_item=False,
                                          _use_hardlink=True)
                progress_cb(len(batch))

    # download images from server
    if len(images_to_download) > 0:
        # prepare lists for api methods
        img_ids = []
        img_paths = []
        for img_info in images_to_download:
            img_ids.append(img_info.id)
            # TODO download to a temp file and use dataset api to add the image to the dataset.
            img_paths.append(
                os.path.join(dataset.img_dir, _maybe_append_image_extension(img_info.name, img_info.ext)))

        # download annotations
        ann_info_list = api.annotation.download_batch(dataset_id, img_ids, progress_cb)
        img_name_to_ann = {ann.image_id: ann.annotation for ann in ann_info_list}
        api.image.download_paths(dataset_id, img_ids, img_paths, progress_cb)
        for img_info, img_path in zip(images_to_download, img_paths):
            dataset.add_item_file(img_info.name, img_path, img_name_to_ann[img_info.id])

        if cache:
            img_hashes = [img_info.hash for img_info in images_to_download]
            cache.write_objects(img_paths, img_hashes)


# coding: utf-8

import os
import struct
import requests
import traceback
import time

from supervisely_lib.worker_api.retriers import retriers_from_cfg
from supervisely_lib.io.network_exceptions import process_requests_exception, process_unhandled_request


class AgentAPI:
    def __init__(self, token, server_address, ext_logger, cfg_path=None):
        self.logger = ext_logger
        self.server_address = server_address
        if ('http://' not in self.server_address) and ('https://' not in self.server_address):
            self.server_address = os.path.join('http://', self.server_address)
        self.server_address = os.path.join(self.server_address, 'agent')
        self.headers = {
            'Content-type': 'application/octet-stream',
            'Accept-Encoding': 'deflate',  # to override default 'Accept-Encoding': 'gzip, deflate'
        }
        if token is not None:
            self.headers['x-token'] = token
        self._retriers = retriers_from_cfg(cfg_path)

    def _get_retrier(self, api_method_name, common_selector):
        retrier = self._retriers.get(api_method_name)
        if retrier is None:
            retrier = self._retriers[common_selector]
        return retrier

    def add_to_metadata(self, key, value):
        self.headers[key] = value

    def rm_from_metadata(self, key):
        self.headers.pop(key, None)

    def _send_request(self, api_method_name, request_data, timeout, in_stream, addit_headers):
        url = os.path.join(self.server_address, api_method_name)
        if not addit_headers:
            addit_headers = {}
        cur_header = {**self.headers, **addit_headers}
        not_log_request = api_method_name != 'Log'
        server_reply = None
        try:
            server_reply = requests.post(url, headers=cur_header, data=request_data, stream=in_stream, timeout=timeout)
            server_reply.raise_for_status()
        except requests.RequestException as exc:
            process_requests_exception(self.logger, exc, api_method_name, url,
                                       verbose=not_log_request, swallow_exc=False, response=server_reply)
        except Exception as exc:
            process_unhandled_request(self.logger, exc)

        return server_reply

    # magic value 4 means four bytes for message length
    # http://www.sureshjoshi.com/development/streaming-protocol-buffers/
    # https://www.datadoghq.com/blog/engineering/protobuf-parsing-in-python/
    def _get_input_stream(self, api_method_name, res_proto_fn, request_data, timeout, addit_headers):
        def cut_len(msg_buff_):
            cur_m_len = struct.unpack('>I', msg_buff_[0:4])[0]
            return cur_m_len, msg_buff_[4:]

        def append_to_msg_buffer(msg_len_, msg_buf_, rest_buf):
            if msg_len_ > len(msg_buf_) + len(rest_buf):
                msg_buf_ = msg_buf_ + rest_buf
                rest_buf = b""
            else:
                tmp_cut_len = msg_len_ - len(msg_buf_)
                msg_buf_ = msg_buf_ + rest_buf[0:tmp_cut_len]
                rest_buf = rest_buf[tmp_cut_len:]
            return msg_buf_, rest_buf

        # request package crashes on empty chunks
        try:
            with self._send_request(api_method_name, request_data, timeout,
                                    in_stream=True, addit_headers=addit_headers) as reply:
                msg_len = None
                msg_buf = b""
                for buffer in reply.iter_content(chunk_size=None):
                    while len(buffer) > 0:
                        if msg_len is None:
                            if len(msg_buf) < 4:
                                msg_buf, buffer = append_to_msg_buffer(4, msg_buf, buffer)
                            if len(msg_buf) < 4:
                                continue
                            msg_len, msg_buf = cut_len(msg_buf)
                        msg_buf, buffer = append_to_msg_buffer(msg_len, msg_buf, buffer)
                        if msg_len == len(msg_buf):
                            if msg_len == 0:
                                pass
                            else:
                                proto_msg = res_proto_fn()
                                proto_msg.ParseFromString(msg_buf)
                                yield proto_msg
                            msg_len = None
                            msg_buf = b""
                if msg_len is not None:
                    raise RuntimeError('MISSED_STREAM_CHUNKS')
        except requests.exceptions.ChunkedEncodingError:
            raise RuntimeError('Unknown error during stream. Please contact support.')
        except requests.RequestException:
            raise
        except Exception as e:
            self.logger.error(traceback.format_exc(), exc_info=True, extra={'exc_str': str(e)})
            raise e

    def _put_out_stream(self, api_method_name, res_proto_fn, chunk_generator, timeout, addit_headers):
        def bindata_generator():
            for chunk in chunk_generator:
                size = chunk.ByteSize()
                res_bytes_with_len = struct.pack('>I', size) + chunk.SerializeToString()
                yield res_bytes_with_len

        resp = self._send_request(api_method_name, bindata_generator(), timeout,
                                  in_stream=False, addit_headers=addit_headers)
        res_proto = res_proto_fn()
        res_proto.ParseFromString(resp.content)
        return res_proto

    # will not log it now
    def simple_request(self, api_method_name, res_proto_fn, proto_request, addit_headers=None):
        data_to_send = proto_request.SerializeToString()
        retrier = self._get_retrier(api_method_name, '__simple_request')
        resp = retrier.request(self._send_request, api_method_name, data_to_send, in_stream=False, addit_headers=addit_headers)
        if resp is None:
            return None  # swallowed exception
        res_proto = res_proto_fn()
        res_proto.ParseFromString(resp.content)
        return res_proto

    def get_stream_with_data(self, api_method_name, res_proto_fn, proto_request, addit_headers=None):
        data_to_send = proto_request.SerializeToString()
        retrier = self._get_retrier(api_method_name, '__data_stream_in')
        yield from retrier.request(self._get_input_stream,
                                   api_method_name, res_proto_fn, data_to_send, addit_headers=addit_headers)

    def get_endless_stream(self, api_method_name, res_proto_fn, proto_request, addit_headers=None,
                           server_fail_limit=10, wait_server_sec=10):
        data_to_send = proto_request.SerializeToString()
        for attempt in range(server_fail_limit):
            retrier = self._get_retrier(api_method_name, '__endless_stream_in')
            yield from retrier.request(self._get_input_stream,
                                       api_method_name, res_proto_fn, data_to_send, addit_headers=addit_headers)
            self.logger.warn('Endless input stream end', extra={'method': api_method_name})
            if attempt != server_fail_limit - 1:
                time.sleep(wait_server_sec)

    def put_stream_with_data(self, api_method_name, res_proto_fn, chunk_generator, addit_headers=None):
        retrier = self._get_retrier(api_method_name, '__data_stream_out')
        res = retrier.request(self._put_out_stream,
                              api_method_name, res_proto_fn, chunk_generator, addit_headers=addit_headers)
        return res

    @classmethod
    def catch_http_err(cls, code_list, fn, *args, **kwargs):
        try:
            res = fn(*args, **kwargs)
            return res, None
        except requests.exceptions.HTTPError as e:
            res_code = e.response.status_code
            if res_code in code_list:
                return None, res_code
            raise
